## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import os
import sys
import logging
import shlex
import functools
import json
import semantic_version

from . import uhf
uhf(__name__)

from . import center, util, clibs, integration, api, web, repo, launcher, runner
from .qt import QtCore, QtWidgets, load_styles
from .ui.hell import Ui_MainWindow as Ui_Hell
from .ui.gogextract import Ui_Dialog as Ui_Gogextract
from .ui.add_repo import Ui_Dialog as Ui_AddRepo
from .ui.settings2 import Ui_Dialog as Ui_Settings
from .ui.settings_about import Ui_Form as Ui_Settings_About
from .ui.settings_sources import Ui_Form as Ui_Settings_Sources
from .ui.settings_versions import Ui_Form as Ui_Settings_Versions
from .ui.settings_knossos import Ui_Form as Ui_Settings_Knossos
from .ui.settings_fso import Ui_Form as Ui_Settings_Fso
from .ui.settings_video import Ui_Form as Ui_Settings_Video
from .ui.settings_audio import Ui_Form as Ui_Settings_Audio
from .ui.settings_input import Ui_Form as Ui_Settings_Input
from .ui.settings_network import Ui_Form as Ui_Settings_Network
from .ui.settings_help import Ui_Form as Ui_Settings_Help
from .ui.flags import Ui_Dialog as Ui_Flags
from .ui.install import Ui_Dialog as Ui_Install
from .ui.mod_settings import Ui_Dialog as Ui_Mod_Settings
from .ui.mod_versions import Ui_Dialog as Ui_Mod_Versions
from .ui.log_viewer import Ui_Dialog as Ui_Log_Viewer
from .tasks import run_task, GOGExtractTask, InstallTask, UninstallTask, WindowsUpdateTask, CheckTask, CheckFilesTask

# Keep references to all open windows to prevent the GC from deleting them.
_open_wins = []


class QDialog(QtWidgets.QDialog):

    def __init__(self, *args):
        super(QDialog, self).__init__(*args)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)


class QMainWindow(QtWidgets.QMainWindow):

    def __init__(self, *args):
        super(QMainWindow, self).__init__(*args)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

    def closeEvent(self, e):
        e.accept()

        center.app.quit()

    def changeEvent(self, event):
        if event.type() == QtCore.QEvent.ActivationChange and integration.current is not None:
            integration.current.annoy_user(False)

        return super(QMainWindow, self).changeEvent(event)


class Window(object):
    win = None
    closed = True
    _is_window = True
    _tpl_cache = None
    _styles = ''
    _qchildren = None

    def __init__(self, window=True):
        super(Window, self).__init__()

        self._is_window = window
        self._tpl_cache = {}
        self._qchildren = []

    def _create_win(self, ui_class, qt_widget=QDialog):
        if not self._is_window:
            qt_widget = QtWidgets.QWidget

        self.win = util.init_ui(ui_class(), qt_widget())

    def open(self):
        global _open_wins

        if not self.closed:
            return

        self.closed = False
        _open_wins.append(self)
        self.win.destroyed.connect(self._del)
        self.win.show()

    def close(self):
        self.win.close()

    def _del(self):
        global _open_wins

        self.closed = True

        if self in _open_wins:
            _open_wins.remove(self)

    def label_tpl(self, label, **vars):
        name = label.objectName()
        if name in self._tpl_cache:
            text = self._tpl_cache[name]
        else:
            text = self._tpl_cache[name] = label.text()

        for name, value in vars.items():
            text = text.replace('{' + name + '}', value)

        label.setText(text)

    def load_styles(self, path, append=True):
        data = load_styles(path)

        if append:
            self._styles += data
        else:
            self._styles = data

        self.win.setStyleSheet(self._styles)


class HellWindow(Window):
    _tasks = None
    _mod_filter = 'installed'
    browser_ctrl = None
    progress_win = None

    def __init__(self, window=True):
        super(HellWindow, self).__init__(window)
        self._tasks = {}

        self._create_win(Ui_Hell, QMainWindow)
        self.browser_ctrl = web.BrowserCtrl(self.win.webView)

        self.win.backButton.clicked.connect(self.win.webView.back)
        self.win.searchEdit.textEdited.connect(self.search_mods)
        self.win.updateButton.clicked.connect(api.fetch_list)
        self.win.settingsButton.clicked.connect(self.show_settings)

        for button in ('installed', 'available', 'updates', 'progress'):
            bt = getattr(self.win, button + 'Button')
            hdl = functools.partial(self.update_mod_buttons, button)

            setattr(self, '_mod_button_hdl_' + button, hdl)
            bt.clicked.connect(hdl)

        self.win.webView.loadStarted.connect(self.show_indicator)
        self.win.webView.loadFinished.connect(self.check_loaded)

        center.signals.update_avail.connect(self.ask_update)
        center.signals.repo_updated.connect(self.check_new_repo)
        center.signals.task_launched.connect(self.watch_task)

        self.win.pageControls.hide()
        self.win.progressInfo.hide()
        self.open()

    def _del(self):
        center.signals.update_avail.disconnect(self.ask_update)
        center.signals.repo_updated.disconnect(self.check_new_repo)
        center.signals.task_launched.disconnect(self.watch_task)

        super(HellWindow, self)._del()

    def check_fso(self, interactive=True):
        if center.settings['fs2_path'] is not None:
            self.win.pageControls.setEnabled(True)
            self.win.updateButton.setEnabled(True)
            self.win.searchEdit.setEnabled(True)
            self.win.tabButtons.setEnabled(True)

            if interactive and self.win.webView.url().toString() == 'qrc:///html/welcome.html':
                self.update_mod_buttons('available')
        else:
            # Make sure the user has a complete configuration
            if not SettingsWindow.has_config():
                tmp = SettingsWindow()
                tmp.write_config()
                tmp.close()

            self.win.webView.load(QtCore.QUrl('qrc:///html/welcome.html'))
            self.win.pageControls.setEnabled(False)
            self.win.updateButton.setEnabled(False)
            self.win.searchEdit.setEnabled(False)
            self.win.tabButtons.setEnabled(False)

    def check_new_repo(self):
        updates = len(center.installed.get_updates())
        self.win.updatesButton.setText('Updates (%d)' % updates)

        self.update_mod_list()

    def ask_update(self, version):
        # We only have an updater for windows.
        if sys.platform.startswith('win'):
            msg = 'There\'s an update available!\nDo you want to install Knossos %s now?' % str(version)
            buttons = QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
            result = QtWidgets.QMessageBox.question(center.app.activeWindow(), 'Knossos', msg, buttons)
            if result == QtWidgets.QMessageBox.Yes:
                run_task(WindowsUpdateTask())
        else:
            msg = 'There\'s an update available!\nYou should update to Knossos %s.' % str(version)
            QtWidgets.QMessageBox.information(center.app.activeWindow(), 'Knossos', msg)

    def update_repo_list(self):
        api.fetch_list()

    def search_mods(self):
        mods = None

        if self._mod_filter == 'installed':
            mods = center.installed.mods
        elif self._mod_filter == 'available':
            mods = {}
            for mid, mvs in center.mods.mods.items():
                if mid not in center.installed.mods:
                    mods[mid] = mvs
        elif self._mod_filter == 'updates':
            mods = {}
            for mid in center.installed.get_updates():
                mods[mid] = center.installed.mods[mid]
        elif self._mod_filter == 'progress':
            mods = {}

        # Now filter the mods.
        query = self.win.searchEdit.text().lower()
        result = {}
        for mid, mvs in mods.items():
            if query in mvs[0].title.lower():
                result[mid] = [mod.get() for mod in mvs]

        return result, self._mod_filter

    def update_mod_list(self):
        result, filter_ = self.search_mods()
        self.browser_ctrl.get_bridge().updateModlist.emit(result, filter_)

    def show_settings(self):
        SettingsWindow()

    def show_indicator(self):
        self.win.setCursor(QtCore.Qt.BusyCursor)

    def check_loaded(self, success):
        self.win.unsetCursor()

        page = self.win.webView.url().toString()
        if page == 'qrc:///html/modlist.html':
            self.win.listControls.show()
            self.win.pageControls.hide()

            # QtCore.QTimer.singleShot(30, lambda: self.update_mod_buttons(self._mod_filter))
        else:
            self.win.listControls.hide()
            self.win.pageControls.show()

    def update_mod_buttons(self, clicked=None):
        for button in ('installed', 'available', 'updates', 'progress'):
            getattr(self.win, button + 'Button').setChecked(button == clicked)

        self._mod_filter = clicked
        page = self.win.webView.url().toString()
        if page != 'qrc:///html/modlist.html':
            self.win.webView.load(QtCore.QUrl('qrc:///html/modlist.html'))
        else:
            self.update_mod_list()

    def watch_task(self, task):
        self._tasks[id(task)] = task
        self.browser_ctrl.get_bridge().taskStarted.emit(id(task), task.title)

        task.done.connect(functools.partial(self._forget_task, task))
        task.progress.connect(functools.partial(self._track_progress, task))

        if len(self._tasks) == 1:
            self.win.progressInfo.show()
            self.win.progressLabel.setText(task.title)
            self.win.progressBar.setValue(0)

            integration.current.show_progress(0)
        else:
            # TODO: Stop being lazy and calculate the aggregate progress.
            self.win.progressBar.hide()
            self.win.progressLabel.setText('Working...')

    def _track_progress(self, task, pi):
        subs = [item for item in pi[1].values()]
        self.browser_ctrl.get_bridge().taskProgress.emit(id(task), pi[0] * 100, json.dumps(subs), pi[2])

        if len(self._tasks) == 1:
            integration.current.set_progress(pi[0])
            self.win.progressBar.setValue(pi[0] * 100)

    def _forget_task(self, task):
        logging.debug('Task "%s" (%d) finished.', task.title, id(task))
        self.browser_ctrl.get_bridge().taskFinished.emit(id(task))
        del self._tasks[id(task)]

        if len(self._tasks) == 1:
            task = list(self._tasks.values())[0]
            self.win.progressLabel.setText(task.title)
            self.win.progressBar.setValue(task.get_progress()[0])
            self.win.progressBar.show()
        elif len(self._tasks) == 0:
            self.win.progressInfo.hide()
            integration.current.hide_progress()

    def abort_task(self, task):
        if task in self._tasks:
            self._tasks[task].abort()


# TODO: Move the settings read/write logic into another module.
class SettingsWindow(Window):
    _tabs = None
    _builds = None

    def __init__(self):
        self._tabs = {}
        self._create_win(Ui_Settings)

        self.win.treeWidget.expandAll()
        self.win.treeWidget.currentItemChanged.connect(self.select_tab)
        self.win.saveButton.clicked.connect(self.write_config)

        self._tabs['About Knossos'] = tab = util.init_ui(Ui_Settings_About(), QtWidgets.QWidget())

        self._tabs['Launcher settings'] = tab = util.init_ui(Ui_Settings_Knossos(), QtWidgets.QWidget())
        tab.versionLabel.setText(center.VERSION)
        tab.maxDownloads.setValue(center.settings['max_downloads'])

        if launcher.log_path is None:
            tab.debugLog.hide()
        else:
            tab.debugLog.clicked.connect(self.show_knossos_log)

        tab.clearHashes.clicked.connect(self.clear_hash_cache)

        if center.settings['update_channel'] == 'stable':
            tab.updateChannel.setCurrentIndex(0)
        else:
            tab.updateChannel.setCurrentIndex(1)

        if center.settings['update_notify']:
            tab.updateNotify.setCheckState(QtCore.Qt.Checked)
        else:
            tab.updateNotify.setCheckState(QtCore.Qt.Unchecked)

        if center.settings['use_raven']:
            tab.reportErrors.setCheckState(QtCore.Qt.Checked)
        else:
            tab.reportErrors.setCheckState(QtCore.Qt.Unchecked)

        tab.maxDownloads.valueChanged.connect(self.update_max_downloads)
        tab.updateChannel.currentIndexChanged.connect(self.save_update_settings)
        tab.updateNotify.stateChanged.connect(self.save_update_settings)
        tab.reportErrors.stateChanged.connect(self.save_report_settings)

        self._tabs['Retail install'] = tab = GogExtractWindow(False)
        tab.win.cancelButton.hide()

        self._tabs['Mod sources'] = tab = util.init_ui(Ui_Settings_Sources(), QtWidgets.QWidget())
        tab.addSource.clicked.connect(self.add_repo)
        tab.editSource.clicked.connect(self.edit_repo)
        tab.removeSource.clicked.connect(self.remove_repo)
        tab.sourceList.itemDoubleClicked.connect(self.edit_repo)

        self._tabs['Mod versions'] = tab = util.init_ui(Ui_Settings_Versions(), QtWidgets.QWidget())

        self._tabs['Game settings'] = tab = util.init_ui(Ui_Settings_Fso(), QtWidgets.QWidget())
        tab.browseButton.clicked.connect(self.select_fs2_path)
        tab.build.activated.connect(self.save_build)
        tab.openLog.clicked.connect(self.show_fso_log)

        self._tabs['Video'] = tab = util.init_ui(Ui_Settings_Video(), QtWidgets.QWidget())
        self._tabs['Audio'] = tab = util.init_ui(Ui_Settings_Audio(), QtWidgets.QWidget())
        self._tabs['Input'] = tab = util.init_ui(Ui_Settings_Input(), QtWidgets.QWidget())
        self._tabs['Network'] = tab = util.init_ui(Ui_Settings_Network(), QtWidgets.QWidget())

        self._tabs['Default flags'] = tab = FlagsWindow(window=False)
        if center.settings['fs2_path'] is None:
            tab.win.setEnabled(False)

        self._tabs['Help'] = util.init_ui(Ui_Settings_Help(), QtWidgets.QWidget())

        center.signals.fs2_path_changed.connect(self.read_config)
        center.signals.fs2_bin_changed.connect(tab.read_flags)

        self.update_repo_list()
        self.read_config()

        self.show_tab('About Knossos')
        self.open()

    def _del(self):
        center.signals.fs2_path_changed.disconnect(self.read_config)
        center.signals.fs2_bin_changed.disconnect(self._tabs['Default flags'].read_flags)

        super(SettingsWindow, self)._del()

    def show_tab(self, tab):
        if tab not in self._tabs:
            logging.error('SettingsWindow: Tab "%s" not found!', tab)
            return

        self.win.layout.removeWidget(self.win.currentTab)
        self.win.currentTab.hide()
        self.win.currentTab = getattr(self._tabs[tab], 'win', self._tabs[tab])
        self.win.layout.addWidget(self.win.currentTab)
        self.win.currentTab.show()

    def select_tab(self, item, prev):
        self.show_tab(item.text(0))

    def update_repo_list(self):
        tab = self._tabs['Mod sources']
        tab.sourceList.clear()

        for i, r in enumerate(center.settings['repos']):
            item = QtWidgets.QListWidgetItem(r[1], tab.sourceList)
            item.setData(QtCore.Qt.UserRole, i)

    def _edit_repo(self, repo_=None, idx=None):
        win = util.init_ui(Ui_AddRepo(), QDialog(self.win))
        win.setAttribute(QtCore.Qt.WA_DeleteOnClose, False)

        win.okButton.clicked.connect(win.accept)
        win.cancelButton.clicked.connect(win.reject)

        if repo_ is not None:
            win.source.setText(repo_[0])
            win.title.setText(repo_[1])

        if win.exec_() == QtWidgets.QMessageBox.Accepted:
            source = win.source.text()
            title = win.title.text()

            if idx is None:
                found = False

                for r_source, r_title in center.settings['repos']:
                    if r_source == source:
                        found = True
                        QtWidgets.QMessageBox.critical(self.win, 'Error', 'This source is already in the list! (As "%s")' % (r_title))
                        break

                if not found:
                    center.settings['repos'].append((source, title))
            else:
                center.settings['repos'][idx] = (source, title)

            api.save_settings()
            self.update_repo_list()

        win.deleteLater()

    def add_repo(self):
        self._edit_repo()

    def edit_repo(self):
        tab = self._tabs['Mod sources']
        item = tab.sourceList.currentItem()
        if item is not None:
            idx = item.data(QtCore.Qt.UserRole)
            self._edit_repo(center.settings['repos'][idx], idx)

    def remove_repo(self):
        tab = self._tabs['Mod sources']
        item = tab.sourceList.currentItem()
        if item is not None:
            idx = item.data(QtCore.Qt.UserRole)
            answer = QtWidgets.QMessageBox.question(self.win, 'Are you sure?', 'Do you really want to remove "%s"?' % (item.text()),
                                                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No, QtWidgets.QMessageBox.No)

            if answer == QtWidgets.QMessageBox.Yes:
                del center.settings['repos'][idx]

                api.save_settings()
                self.update_repo_list()

    def reorder_repos(self, parent, s_start, s_end, d_parent, d_row):
        tab = self._tabs['Mod sources']
        repos = []
        for row in range(0, tab.sourceList.count()):
            item = self.win.sourceList.item(row)
            repos.append(center.settings['repos'][item.data(QtCore.Qt.UserRole)])

        center.settings['repos'] = repos
        api.save_settings()

        # NOTE: This call is normally redundant but I want to make sure that
        # the displayed list is always the same as the actual list in settings['repos'].
        # Once this feature is stable this call can be removed.
        self.update_repo_list()

    def update_max_downloads(self, num=None):
        old_num = center.settings['max_downloads']
        center.settings['max_downloads'] = num = self._tabs['Launcher settings'].maxDownloads.value()

        if num != old_num:
            util.DL_POOL.set_capacity(num)
            api.save_settings()

    def get_ratio(self, w, h):
        w = int(w)
        h = int(h)
        ratio = (1.0 * w / h)
        ratio = '{:.1f}'.format(ratio)
        if ratio == '1.3':
            ratio_string = '4:3'
        elif ratio == '1.6':
            ratio_string = '16:10'
        elif ratio == '1.8':
            ratio_string = '16:9'
        else:
            ratio_string = 'custom'
        return ratio_string

    @staticmethod
    def has_config():
        if sys.platform.startswith('win'):
            config = QtCore.QSettings('HKEY_LOCAL_MACHINE\\Software\\Volition\\Freespace2', QtCore.QSettings.NativeFormat)
        else:
            config_file = os.path.join(api.get_fso_profile_path(), 'fs2_open.ini')
            config = QtCore.QSettings(config_file, QtCore.QSettings.IniFormat)

        if not sys.platform.startswith('win'):
            config.beginGroup('Default')

        return config.contains('VideocardFs2open')

    def read_config(self):
        fs2_path = center.settings['fs2_path']
        fs2_bin = center.settings['fs2_bin']

        tab = self._tabs['Game settings']
        tab.fs2PathLabel.setText(fs2_path)
        tab.build.clear()

        if fs2_path is not None:
            self._builds = bins = api.get_executables()

            idx = 0
            for name, path in bins:
                tab.build.addItem(name)

                if path == fs2_bin:
                    tab.build.setCurrentIndex(idx)

                idx += 1

            if len(bins) == 1:
                # Found only one binary, select it by default.
                tab.build.setEnabled(False)
                self.save_build()

        # ---Read fs2_open.ini or the registry---
        if sys.platform.startswith('win'):
            self.config = config = QtCore.QSettings('HKEY_LOCAL_MACHINE\\Software\\Volition\\Freespace2', QtCore.QSettings.NativeFormat)
        else:
            config_file = os.path.join(api.get_fso_profile_path(), 'fs2_open.ini')
            self.config = config = QtCore.QSettings(config_file, QtCore.QSettings.IniFormat)

        # Be careful with any change, the keys are all case sensitive.
        if not sys.platform.startswith('win'):
            config.beginGroup('Default')

        # video settings
        if config.contains('VideocardFs2open'):
            rawres = config.value('VideocardFs2open')
            res = rawres.split('(')[1].split(')')[0]
            res_width, res_height = res.split('x')
            depth = rawres.split(')x')[1][0:2]

            try:
                res_width = int(res_width)
                res_height = int(res_height)
            except TypeError:
                res_width = res_height = None
        else:
            res_width = None
            res_height = None
            depth = None

        texfilter = config.value('TextureFilter', None)
        af = config.value('OGL_AnisotropicFilter', None)
        aa = config.value('OGL_AntiAliasSamples', None)

        # joysticks
        joystick_id = config.value('CurrentJoystick', None)

        # network settings
        net_connection = config.value('NetworkConnection', None)
        net_speed = config.value('ConnectionSpeed', None)
        net_port = config.value('ForcePort', None)
        if not sys.platform.startswith('win'):
            config.endGroup()

        net_ip = config.value('Network/CustomIP', None)

        config.beginGroup('Sound')

        # sound settings
        playback_device = config.value('PlaybackDevice', None)
        capture_device = config.value('CaptureDevice', None)
        enable_efx = config.value('EnableEFX', None)
        sample_rate = config.value('SampleRate', None)

        config.endGroup()

        # ---Video settings---
        # Screen resolution
        vid_res = self._tabs['Video'].resolution
        vid_res.clear()

        raw_modes = clibs.get_modes()
        modes = []
        for mode in raw_modes:
            if mode not in modes and not (mode[0] < 800 or mode[1] < 600):
                modes.append(mode)

        for i, (width, height) in enumerate(modes):
            vid_res.addItem('{0} x {1} ({2})'.format(width, height, self.get_ratio(width, height)))

            if width == res_width and height == res_height:
                vid_res.setCurrentIndex(i)

        # Screen depth
        vid_depth = self._tabs['Video'].colorDepth
        vid_depth.clear()
        vid_depth.addItems(['32-bit', '16-bit'])

        index = vid_depth.findText('{0}-bit'.format(depth))
        if index != -1:
            vid_depth.setCurrentIndex(index)

        # Texture filter
        vid_texfilter = self._tabs['Video'].textureFilter
        vid_texfilter.clear()
        vid_texfilter.addItems(['Bilinear', 'Trilinear'])

        try:
            index = int(texfilter)
        except TypeError:
            index = 0

        # If the SCP adds a new texture filder, we should change this part.
        if index > 1:
            index = 0
        vid_texfilter.setCurrentIndex(index)

        # Antialiasing
        vid_aa = self._tabs['Video'].antialiasing
        vid_aa.clear()
        vid_aa.addItems(['Off', '2x', '4x', '8x', '16x'])

        index = vid_aa.findText('{0}x'.format(aa))
        if index == -1:
            index = 0
        vid_aa.setCurrentIndex(index)

        # Anisotropic filtering
        vid_af = self._tabs['Video'].anisotropic
        vid_af.clear()
        vid_af.addItems(['Off', '1x', '2x', '4x', '8x', '16x'])

        index = vid_af.findText('{0}x'.format(af))
        if index == -1:
            index = 0
        vid_af.setCurrentIndex(index)

        # ---Sound settings---
        if clibs.can_detect_audio():
            snd_devs, snd_default, snd_captures, snd_default_capture = clibs.list_audio_devs()
            snd_playback = self._tabs['Audio'].playbackDevice
            snd_capture = self._tabs['Audio'].captureDevice
            snd_playback.clear()
            snd_capture.clear()

            for name in snd_devs:
                snd_playback.addItem(name)

            if playback_device is not None:
                index = snd_playback.findText(playback_device)
                if index == -1:
                    index = snd_playback.findText(snd_default)

            if index != -1:
                snd_playback.setCurrentIndex(index)

            # Fill input device combobox:
            for name in snd_captures:
                snd_capture.addItem(name)

            if capture_device is not None:
                index = snd_capture.findText(capture_device)
                if index == -1:
                    index = snd_capture.findText(snd_default_capture)

            if index != -1:
                snd_capture.setCurrentIndex(index)

            # Fill sample rate textbox :
            snd_samplerate = self._tabs['Audio'].sampleRate
            if util.is_number(sample_rate):
                sample_rate = int(sample_rate)
                if sample_rate > 0 and sample_rate < 1000000:
                    snd_samplerate.setValue(sample_rate)
                else:
                    snd_samplerate.setValue(44100)
            else:
                snd_samplerate.setValue(44100)

            # Fill EFX checkbox :
            if enable_efx == '1':
                self._tabs['Audio'].enableEFX.setChecked(True)
            else:
                self._tabs['Audio'].enableEFX.setChecked(False)

        # ---Joystick settings---
        joysticks = clibs.list_joysticks()
        ctrl_joystick = self._tabs['Input'].controller

        ctrl_joystick.clear()
        ctrl_joystick.addItem('No Joystick')
        for joystick in joysticks:
            ctrl_joystick.addItem(joystick)

        if util.is_number(joystick_id):
            if joystick_id == '99999':
                ctrl_joystick.setCurrentIndex(0)
            else:
                ctrl_joystick.setCurrentIndex(int(joystick_id) + 1)
                if ctrl_joystick.currentText() == '':
                    ctrl_joystick.setCurrentIndex(0)
        else:
            ctrl_joystick.setCurrentIndex(0)

        if len(joysticks) == 0:
            ctrl_joystick.setEnabled(False)

        # ---Keyboard settings---
        kls = self._tabs['Input'].keyLayout
        if sys.platform.startswith('linux'):
            kls.clear()
            for i, layout in enumerate(('default (qwerty)', 'qwertz', 'azerty')):
                kls.addItem(layout)
                if layout == center.settings['keyboard_layout']:
                    kls.setCurrentIndex(i)

            if center.settings['keyboard_setxkbmap']:
                self._tabs['Input'].useSetxkbmap.setChecked(True)
            else:
                self._tabs['Input'].useSetxkbmap.setChecked(False)
        else:
            kls.setDisabled(True)
            self._tabs['Input'].useSetxkbmap.setDisabled(True)

        # ---Network settings---
        net_type = self._tabs['Network'].connectionType
        net_speed = self._tabs['Network'].connectionSpeed
        net_ip_f = self._tabs['Network'].forceAddress
        net_port_f = self._tabs['Network'].localPort

        net_type.clear()
        net_type.addItems(['None', 'Dialup', 'Broadband/LAN'])
        net_connections_read = {'none': 0, 'dialup': 1, 'LAN': 2}
        if net_connection in net_connections_read:
            index = net_connections_read[net_connection]
        else:
            index = 2

        net_type.setCurrentIndex(index)

        net_speed.clear()
        net_speed.addItems(['None', '28k modem', '56k modem', 'ISDN', 'DSL', 'Cable/LAN'])
        net_speeds_read = {'none': 0, 'Slow': 1, '56K': 2, 'ISDN': 3, 'Cable': 4, 'Fast': 5}
        if net_speed in net_speeds_read:
            index = net_speeds_read[net_speed]
        else:
            index = 5

        net_speed.setCurrentIndex(index)
        net_ip_f.setText(net_ip)
        net_port_f.setText(net_port)

    def write_config(self):
        config = self.config

        if sys.platform.startswith('win'):
            section = ''
        else:
            config.beginGroup('Default')
            section = 'Default/'

        # Getting ready to write key=value pairs to the ini file
        # Set video
        new_res_width, new_res_height = self._tabs['Video'].resolution.currentText().split(' (')[0].split(' x ')
        new_depth = self._tabs['Video'].colorDepth.currentText().split('-')[0]
        new_res = 'OGL -({0}x{1})x{2} bit'.format(new_res_width, new_res_height, new_depth)
        config.setValue('VideocardFs2open', new_res)

        new_texfilter = self._tabs['Video'].textureFilter.currentIndex()
        config.setValue('TextureFilter', new_texfilter)

        new_aa = self._tabs['Video'].antialiasing.currentText().split('x')[0]
        config.setValue('OGL_AntiAliasSamples', new_aa)

        new_af = self._tabs['Video'].anisotropic.currentText().split('x')[0]
        config.setValue('OGL_AnisotropicFilter', new_af)

        if not sys.platform.startswith('win'):
            config.endGroup()

        # sound
        new_playback_device = self._tabs['Audio'].playbackDevice.currentText()
        # ^ wxlauncher uses the same string as CaptureDevice, instead of what openal identifies as the playback device ? Why ?
        # ^ So I do it the way openal is supposed to work, but I'm not sure FS2 really behaves that way
        config.setValue('Sound/PlaybackDevice', new_playback_device)
        config.setValue(section + 'SoundDeviceOAL', new_playback_device)
        # ^ Useless according to SCP members, but wxlauncher does it anyway

        new_capture_device = self._tabs['Audio'].captureDevice.currentText()
        config.setValue('Sound/CaptureDevice', new_capture_device)

        if self._tabs['Audio'].enableEFX.isChecked() is True:
            new_enable_efx = 1
        else:
            new_enable_efx = 0
        config.setValue('Sound/EnableEFX', new_enable_efx)

        new_sample_rate = self._tabs['Audio'].sampleRate.value()
        config.setValue('Sound/SampleRate', new_sample_rate)

        # joystick
        if self._tabs['Input'].controller.currentText() == 'No Joystick':
            new_joystick_id = 99999
        else:
            new_joystick_id = self._tabs['Input'].controller.currentIndex() - 1
        config.setValue(section + 'CurrentJoystick', new_joystick_id)

        # keyboard
        if sys.platform.startswith('linux'):
            key_layout = self._tabs['Input'].keyLayout.currentIndex()
            if key_layout == 0:
                key_layout = 'default'
            else:
                key_layout = self._tabs['Input'].keyLayout.itemText(key_layout)

            center.settings['keyboard_layout'] = key_layout
            center.settings['keyboard_setxkbmap'] = self._tabs['Input'].useSetxkbmap.isChecked()

        # networking
        net_types = {0: 'none', 1: 'dialup', 2: 'LAN'}
        new_net_connection = net_types[self._tabs['Network'].connectionType.currentIndex()]
        config.setValue(section + 'NetworkConnection', new_net_connection)

        net_speeds = {0: 'none', 1: 'Slow', 2: '56K', 3: 'ISDN', 4: 'Cable', 5: 'Fast'}
        new_net_speed = net_speeds[self._tabs['Network'].connectionSpeed.currentIndex()]
        config.setValue(section + 'ConnectionSpeed', new_net_speed)

        new_net_ip = self._tabs['Network'].forceAddress.text()
        if new_net_ip == '...':
            new_net_ip = ''

        if new_net_ip == '':
            config.remove('Network/CustomIP')
        else:
            config.setValue('Network/CustomIP', new_net_ip)

        new_net_port = self._tabs['Network'].localPort.text()
        if new_net_port == '0':
            new_net_port = ''

        if new_net_port == '':
            config.remove(section + 'ForcePort')
        else:
            config.setValue(section + 'ForcePort', int(new_net_port))

        # Save the new configuration.
        config.sync()
        self._tabs['Default flags'].save()
        api.save_settings()

    def select_fs2_path(self):
        api.select_fs2_path()

    def save_build(self):
        fs2_bin = self._builds[self._tabs['Game settings'].build.currentIndex()][1]
        if not os.path.isfile(os.path.join(center.settings['fs2_path'], fs2_bin)):
            return

        old_bin = center.settings['fs2_bin']
        center.settings['fs2_bin'] = fs2_bin

        if not api.get_fso_flags():
            # We failed to run FSO but why?
            rc = runner.run_fs2_silent(['-help'])
            if rc == -128:
                msg = 'The FSO binary "%s" is missing!' % fs2_bin
            elif rc == -127:
                # TODO: At this point we have run ldd twice already and the next call will run it again. Is there any way to avoid this?
                _, missing = runner.fix_missing_libs(os.path.join(center.settings['fs2_path'], fs2_bin))
                msg = 'The FSO binary "%s" is missing %s!' % (fs2_bin, util.human_list(missing))
            else:
                msg = 'The FSO binary quit with code %d!' % rc

            center.settings['fs2_bin'] = old_bin
            QtWidgets.QMessageBox.critical(None, 'Knossos', msg)
            self.read_config()
        else:
            api.save_settings()
            center.signals.fs2_bin_changed.emit()

    def save_update_settings(self, p=None):
        tab = self._tabs['Launcher settings']
        center.settings['update_channel'] = tab.updateChannel.currentText()
        center.settings['update_notify'] = tab.updateNotify.checkState() == QtCore.Qt.Checked
        api.save_settings()

    def save_report_settings(self, p=None):
        tab = self._tabs['Launcher settings']
        center.settings['use_raven'] = tab.reportErrors.checkState() == QtCore.Qt.Checked

        if center.settings['use_raven']:
            api.enable_raven()
        else:
            center.raven = None

        api.save_settings()

    def show_fso_log(self):
        logpath = os.path.join(api.get_fso_profile_path(), 'data/fs2_open.log')

        if not os.path.isfile(logpath):
            QtWidgets.QMessageBox.warning(None, 'Knossos', 'Sorry, but I can\'t find the fs2_open.log file.\nDid you run the debug build?')
        else:
            LogViewer(logpath)

    def show_knossos_log(self):
        LogViewer(launcher.log_path)

    def clear_hash_cache(self):
        util.HASH_CACHE = dict()
        run_task(CheckTask())
        QtWidgets.QMessageBox.information(None, 'Knossos', 'Done!')


class GogExtractWindow(Window):

    def __init__(self, window=True):
        super(GogExtractWindow, self).__init__(window)

        self._create_win(Ui_Gogextract)

        self.win.gogPath.textChanged.connect(self.validate)
        self.win.destPath.textChanged.connect(self.validate)

        self.win.gogButton.clicked.connect(self.select_installer)
        self.win.destButton.clicked.connect(self.select_dest)
        self.win.cancelButton.clicked.connect(self.win.close)
        self.win.installButton.clicked.connect(self.do_install)

        if window:
            self.win.setModal(True)
            self.open()

    def select_installer(self):
        path = QtWidgets.QFileDialog.getOpenFileName(self.win, 'Please select the setup_freespace2_*.exe file.',
                                                 os.path.expanduser('~/Downloads'), 'Executable (*.exe)')
        if isinstance(path, tuple):
            path = path[0]

        if path is not None and path != '':
            if not os.path.isfile(path):
                QtWidgets.QMessageBox.critical(self.win, 'Not a file', 'Please select a proper file!')
                return

            self.win.gogPath.setText(os.path.abspath(path))

    def select_dest(self):
        path = QtWidgets.QFileDialog.getExistingDirectory(self.win, 'Please select the destination directory.', os.path.expanduser('~/Documents'))

        if path is not None and path != '':
            if not os.path.isdir(path):
                QtWidgets.QMessageBox.critical(self.win, 'Not a directory', 'Please select a proper directory!')
                return

            self.win.destPath.setText(os.path.abspath(path))

    def validate(self):
        if os.path.isfile(self.win.gogPath.text()) and os.path.isdir(self.win.destPath.text()):
            self.win.installButton.setEnabled(True)
        else:
            self.win.installButton.setEnabled(False)

    def do_install(self):
        # Just to be sure...
        if os.path.isfile(self.win.gogPath.text()) and os.path.isdir(self.win.destPath.text()):
            center.main_win.update_mod_buttons('progress')

            run_task(GOGExtractTask(self.win.gogPath.text(), self.win.destPath.text()))

            if self._is_window:
                self.close()


class FlagsWindow(Window):
    _flags = None
    _selected = None
    _mod = None
    _mod_arg = ''

    def __init__(self, mod=None, window=True):
        super(FlagsWindow, self).__init__(window)

        self._selected = []
        self._mod = mod

        self._create_win(Ui_Flags)
        if window:
            self.win.setModal(True)

            self.win.okButton.clicked.connect(self.win.accept)
            self.win.cancelButton.clicked.connect(self.win.reject)
            self.win.accepted.connect(self.save)
            self.win.saveButton.hide()
        else:
            self.win.saveButton.clicked.connect(self.save)
            self.win.okButton.hide()
            self.win.cancelButton.hide()

        self.win.easySetup.activated.connect(self._set_easy)
        self.win.easySetup.activated.connect(self.update_display)
        self.win.listType.activated.connect(self._update_list)
        self.win.customFlags.textEdited.connect(self.update_display)
        self.win.flagList.itemClicked.connect(self.update_display)

        self.win.defaultsButton.clicked.connect(self.set_defaults)

        if window:
            self.open()

        if mod is None:
            self.win.defaultsButton.hide()
        else:
            try:
                self._mod_arg = ','.join(mod.get_mod_flag())
            except repo.ModNotFound:
                pass

        self.read_flags()

    def read_flags(self):
        flags = api.get_fso_flags()

        if flags is None:
            self.win.setEnabled(False)
            self.win.cmdLine.setPlainText('Until you select a working FS2 build, I won\'t be able to help you.')
            return

        self.win.setEnabled(True)
        self.win.easySetup.clear()
        self.win.listType.clear()

        for key, name in flags.easy_flags.items():
            self.win.easySetup.addItem(name, key)

        self._flags = flags.flags
        for name in self._flags:
            self.win.listType.addItem(name)

        self.set_selection(api.get_cmdline(self._mod))

    def _set_easy(self):
        if self._flags is None:
            return

        combo = self.win.easySetup
        cur_mode = combo.itemData(combo.currentIndex())

        # Update self._selected.
        self.get_selection()

        for flags in self._flags.values():
            for info in flags:
                if info['on_flags'] & cur_mode == cur_mode and info['name'] not in self._selected:
                    self._selected.append(info['name'])

                if info['off_flags'] & cur_mode == cur_mode and info['name'] in self._selected:
                    self._selected.remove(info['name'])

        self._update_list(save_selection=False)

    def _update_list(self, idx=None, save_selection=True):
        if self._flags is None:
            return

        cur_type = self.win.listType.currentText()

        if save_selection:
            # Remember the currently selected options.
            self.get_selection()

        self.win.flagList.clear()
        for flag in self._flags[cur_type]:
            label = flag['desc']
            if label == '':
                label = flag['name']

            item = QtWidgets.QListWidgetItem(label)
            item.setData(QtCore.Qt.UserRole, flag['name'])
            if flag['name'] in self._selected:
                item.setCheckState(QtCore.Qt.Checked)
            else:
                item.setCheckState(QtCore.Qt.Unchecked)

            self.win.flagList.addItem(item)

    def get_selection(self, split_mod_flag=False):
        count = self.win.flagList.count()

        for i in range(count):
            item = self.win.flagList.item(i)
            flag = item.data(QtCore.Qt.UserRole)
            if item.checkState() == QtCore.Qt.Checked:
                if flag not in self._selected:
                    self._selected.append(flag)
            else:
                if flag in self._selected:
                    self._selected.remove(flag)

        try:
            custom = shlex.split(self.win.customFlags.text())
        except ValueError:
            # Invalid user input...
            custom = []

        if split_mod_flag:
            idx = -1
            for i, flag in enumerate(custom):
                if flag == '-mod':
                    idx = i
                    break

            if idx > -1 and idx + 1 < len(custom):
                mod_arg = ['-mod', custom[idx + 1]]
                custom = custom[:idx] + custom[idx + 2:]
            else:
                mod_arg = []

            return self._selected + custom, mod_arg
        else:
            return self._selected + custom

    def update_display(self):
        fs2_bin = os.path.join(center.settings['fs2_path'], center.settings['fs2_bin'])
        sel, mod_flag = self.get_selection(True)
        if self._mod is not None:
            if len(mod_flag) > 0:
                mod_flag[1] = mod_flag[1].strip(',') + ',' + self._mod_arg
            else:
                mod_flag = ['-mod', self._mod_arg]

        cmdline = ' '.join([fs2_bin] + [shlex.quote(opt) for opt in sel + mod_flag])

        self.win.cmdLine.setPlainText(cmdline)

    def set_selection(self, flags):
        self._selected = flags[:]
        custom = []
        all_flags = []

        for flags in self._flags.values():
            for info in flags:
                all_flags.append(info['name'])

        for opt in self._selected[:]:
            if opt not in all_flags:
                custom.append(shlex.quote(opt))
                self._selected.remove(opt)

        self.win.customFlags.setText(' '.join(custom))
        self._update_list(save_selection=False)
        self.update_display()

    def set_defaults(self):
        cmdlines = center.settings['cmdlines']

        if '#default' in cmdlines:
            self.set_selection(cmdlines['#default'])
        else:
            self.set_selection([])

        self.save()

    def save(self):
        sel = self.get_selection()
        clines = center.settings['cmdlines']

        if self._mod is None:
            clines['#default'] = self.get_selection()
        else:
            mid = self._mod.mid
            if sel == clines['#default']:
                logging.info('Not saving default flags!')
                if mid in clines:
                    del clines[mid]
            else:
                clines[mid] = self.get_selection()

        api.save_settings()


class ModInstallWindow(Window):
    _mod = None
    _pkg_checks = None
    _dep_tpl = None

    def __init__(self, mod, sel_pkgs=[]):
        super(ModInstallWindow, self).__init__()

        self._mod = mod
        self._create_win(Ui_Install)
        self.win.splitter.setStretchFactor(0, 2)
        self.win.splitter.setStretchFactor(1, 1)

        self.label_tpl(self.win.titleLabel, MOD=mod.title)
        self.win.notesField.setPlainText(mod.notes)

        self._pkg_checks = []
        self.show_packages()
        self.update_deps()

        self.win.treeWidget.itemChanged.connect(self.update_selection)

        self.win.accepted.connect(self.install)
        self.win.rejected.connect(self.close)

        self.open()

    def show_packages(self):
        pkgs = self._mod.packages
        try:
            all_pkgs = center.mods.process_pkg_selection(pkgs)
        except repo.ModNotFound as exc:
            logging.exception('Well, I won\'t be installing that...')
            msg = 'I\'m sorry but you won\'t be able to install "%s" because "%s" is missing!' % (self._mod.title, exc.mid)
            QtWidgets.QMessageBox.critical(None, 'Knossos', msg)

            self.close()
            return

        needed_pkgs = self._mod.resolve_deps()
        mods = set()

        for pkg in all_pkgs:
            mods.add(pkg.get_mod())

        # Make sure our current mod comes first.
        if self._mod in mods:
            mods.remove(self._mod)

        mods = [self._mod] + list(mods)
        for mod in mods:
            item = QtWidgets.QTreeWidgetItem(self.win.treeWidget, [mod.title, ''])
            item.setExpanded(True)
            item.setData(0, QtCore.Qt.UserRole, mod)
            c = 0

            for pkg in mod.packages:
                fsize = 0
                for f_item in pkg.files.values():
                    fsize += f_item.get('filesize', 0)

                if fsize > 0:
                    fsize = util.format_bytes(fsize)
                else:
                    fsize = '?'

                sub = QtWidgets.QTreeWidgetItem(item, [pkg.name, fsize])
                is_installed = center.installed.is_installed(pkg)

                if is_installed:
                    sub.setText(1, 'Installed')

                if pkg.status == 'required' or pkg in needed_pkgs or is_installed:
                    sub.setCheckState(0, QtCore.Qt.Checked)
                    sub.setDisabled(True)

                    c += 1
                elif pkg.status == 'recommended':
                    sub.setCheckState(0, QtCore.Qt.Checked)
                    c += 1
                if pkg.status == 'optional':
                    sub.setCheckState(0, QtCore.Qt.Unchecked)

                sub.setData(0, QtCore.Qt.UserRole, pkg)
                sub.setData(0, QtCore.Qt.UserRole + 1, is_installed)
                self._pkg_checks.append(sub)

            if c == len(mod.packages):
                item.setCheckState(0, QtCore.Qt.Checked)
            elif c > 0:
                item.setCheckState(0, QtCore.Qt.PartiallyChecked)
            else:
                item.setCheckState(0, QtCore.Qt.Unchecked)

        self.win.treeWidget.header().resizeSections(QtWidgets.QHeaderView.ResizeToContents)

    def update_selection(self, item, col):
        obj = item.data(0, QtCore.Qt.UserRole)
        if isinstance(obj, repo.Mod):
            cs = item.checkState(0)

            for i in range(item.childCount()):
                child = item.child(i)
                if not child.isDisabled():
                    child.setCheckState(0, cs)

        self.update_deps()

    def get_selected_pkgs(self):
        pkgs = []
        for item in self._pkg_checks:
            if item.checkState(0) == QtCore.Qt.Checked:
                pkgs.append(item.data(0, QtCore.Qt.UserRole))

        pkgs = center.mods.process_pkg_selection(pkgs)
        return pkgs

    def update_deps(self):
        pkg_sel = self.get_selected_pkgs()
        dl_size = 0
        for item in self._pkg_checks:
            pkg = item.data(0, QtCore.Qt.UserRole)

            if not item.data(0, QtCore.Qt.UserRole + 1):
                if item.checkState(0) == QtCore.Qt.Checked or pkg in pkg_sel:
                    for f_item in pkg.files.values():
                        dl_size += f_item.get('filesize', 0)

            if item.isDisabled():
                continue

            if pkg in pkg_sel:
                item.setCheckState(0, QtCore.Qt.Checked)

        self.label_tpl(self.win.dlSizeLabel, DL_SIZE=util.format_bytes(dl_size))

    def install(self):
        center.main_win.update_mod_buttons('progress')

        run_task(InstallTask(self.get_selected_pkgs(), self._mod))
        self.close()


class ModSettingsWindow(Window):
    _mod = None
    _pkg_checks = None
    _mod_versions = None
    _mvs_cbs = None
    _flags = None

    def __init__(self, mod, window=True):
        super(ModSettingsWindow, self).__init__(window)

        self._mod = mod
        self._mod_versions = []
        self._create_win(Ui_Mod_Settings, QDialog)
        self.load_styles(':/ui/themes/default/mod_settings.css')

        self.win.modTitle.setText(self._mod.title)
        self.win.modLogo.setPixmap(QtWidgets.QPixmap(mod.logo_path))
        self.win.modDesc.setPlainText(mod.description)

        lay = QtWidgets.QVBoxLayout(self.win.flagsTab)
        self._flags = FlagsWindow(mod, False)
        lay.addWidget(self._flags.win)

        self._pkg_checks = []
        if isinstance(mod, repo.IniMod):
            self.win.tabWidget.setTabEnabled(2, False)  # Packages
            self.win.tabWidget.setTabEnabled(3, False)  # Versions
            self.win.tabWidget.setTabEnabled(4, False)  # Troubleshooting
        else:
            try:
                rmod = center.mods.query(mod)
            except repo.ModNotFound:
                rmod = mod

            for pkg in rmod.packages:
                p_check = QtWidgets.QCheckBox(pkg.name)
                installed = center.installed.is_installed(pkg)

                if pkg.status == 'required' or installed:
                    p_check.setCheckState(QtCore.Qt.Checked)

                    if pkg.status == 'required':
                        p_check.setDisabled(True)

                        if not installed:
                            p_check.setProperty('error', True)
                            p_check.setText(pkg.name + ' (missing)')
                else:
                    p_check.setCheckState(QtCore.Qt.Unchecked)

                p_check.stateChanged.connect(self.update_dlsize)
                self.win.pkgsLayout.addWidget(p_check)
                self._pkg_checks.append([p_check, installed, pkg])

            self.win.applyPkgChanges.clicked.connect(self.apply_pkg_selection)

            self.win.checkFiles.clicked.connect(self.check_files)
            self.win.delLoose.clicked.connect(self.delete_loose_files)
            self.win.replaceFiles.clicked.connect(self.repair_files)

            self.update_dlsize()
            self.show_versions()

        center.signals.repo_updated.connect(self.update_repo_related)

        self.show_flags_tab()
        self.open()

    def _del(self):
        center.signals.repo_updated.disconnect(self.update_repo_related)

        super(ModSettingsWindow, self)._del()

    def update_repo_related(self):
        try:
            self._mod = center.installed.query(self._mod)
        except repo.ModNotFound:
            self.close()

            # Try to open a new window showing the latest version.
            try:
                m2 = center.installed.query(self._mod.mid)
            except repo.ModNotFound:
                pass
            else:
                ModSettingsWindow(m2)

            return

        self.show_versions()

        for i, item in enumerate(self._pkg_checks):
            checked = item[0].checkState() == QtCore.Qt.Checked
            was_inst = item[1]
            is_inst = center.installed.is_installed(item[2])

            if was_inst != is_inst:
                if checked == was_inst:
                    if is_inst:
                        item[0].setCheckState(QtCore.Qt.Checked)
                    else:
                        item[0].setCheckState(QtCore.Qt.Unchecked)

                if is_inst and item[0].property('error'):
                    item[0].setProperty('error', False)
                    item[0].setText(item[2].name)
                elif not is_inst and item[2].status == 'required' and not item[0].property('error'):
                    item[0].setProperty('error', True)
                    item[0].setText(item[2].name + ' (missing)')

                item[1] = is_inst

        self.update_dlsize()

    def show_flags_tab(self):
        self.win.tabWidget.setCurrentIndex(0)

    def show_pkg_tab(self):
        self.win.tabWidget.setCurrentIndex(1)

    def get_selected_pkgs(self):
        install = []
        remove = []

        for check, is_installed, pkg in self._pkg_checks:
            selected = check.checkState() == QtCore.Qt.Checked
            if selected != is_installed:
                if selected:
                    install.append(pkg)
                else:
                    remove.append(pkg)

        if len(install) > 0:
            try:
                install = center.mods.process_pkg_selection(install)
            except repo.ModNotFound as exc:
                if center.mods.has(self._mod):
                    QtWidgets.QMessageBox.critical(None, 'Knossos', "I'm sorry but I can't install the selected packages because some the dependency \"%s\" is missing!" % exc.mid)
                else:
                    QtWidgets.QMessageBox.critical(None, 'Knossos', "I'm sorry but I can't install new packages for this mod since it's not available anymore!")

                install = []

        remove = list(set(remove) - set(install))

        return install, remove

    def update_dlsize(self):
        install, remove = self.get_selected_pkgs()
        try:
            all_pkgs = center.mods.process_pkg_selection(install)
        except repo.ModNotFound:
            all_pkgs = []

        dl_size = 0

        for pkg in all_pkgs:
            if not center.installed.is_installed(pkg):
                for item in pkg.files.values():
                    dl_size += item.get('filesize', 0)

        self.label_tpl(self.win.dlSizeLabel, DL_SIZE=util.format_bytes(dl_size))

    def apply_pkg_selection(self):
        install, remove = self.get_selected_pkgs()

        if len(remove) == 0 and len(install) == 0:
            # Nothing to do...
            return

        msg = ''
        if len(remove) > 0:
            msg += "I'm going to remove " + util.human_list([p.name for p in remove]) + ".\n"

        if len(install) > 0:
            msg += "I'm going to install " + util.human_list([p.name for p in install if not center.installed.is_installed(p)]) + ".\n"

        box = QtWidgets.QMessageBox()
        box.setIcon(QtWidgets.QMessageBox.Question)
        box.setText(msg)
        box.setInformativeText('Continue?')
        box.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        box.setDefaultButton(QtWidgets.QMessageBox.Yes)

        if box.exec_() == QtWidgets.QMessageBox.Yes:
            run_task(UninstallTask(remove))
            run_task(InstallTask(install))

    def show_versions(self):
        mods = set()
        try:
            for dep in self._mod.resolve_deps():
                mods.add(dep.get_mod())
        except repo.ModNotFound as exc:
            QtWidgets.QMessageBox.critical(None, 'Knossos', 'This mod is missing the dependency "%s"!' % exc.mid)
            return

        mods.add(self._mod)
        self._mvs_cbs = []

        layout = self.win.versionsTab.layout()
        if layout:
            # Clear the current layout
            item = layout.takeAt(0)
            while item:
                w = item.widget()
                if w:
                    w.deleteLater()

                item = layout.takeAt(0)
        else:
            layout = QtWidgets.QGridLayout()
            self.win.versionsTab.setLayout(layout)

        mods = sorted(mods, key=lambda m: m.title)
        for i, mod in enumerate(mods):
            versions = list(center.installed.query_all(mod.mid))
            label = QtWidgets.QLabel(mod.title + ': ')
            label.setWordWrap(True)

            sel = QtWidgets.QComboBox()
            sel.addItem('Latest (%s)' % versions[0].version, None)

            pin = center.installed.get_pin(mod)
            for n, mv in enumerate(versions):
                sel.addItem(str(mv.version), mv.version)

                if pin == mv.version:
                    sel.setCurrentIndex(n + 1)

            editBut = QtWidgets.QPushButton('Edit')

            layout.addWidget(label, i, 0)
            layout.addWidget(sel, i, 1)
            layout.addWidget(editBut, i, 2)

            if mod == self._mod:
                cb_sel = functools.partial(self.update_versions, sel, mod)
                cb_edit = functools.partial(self.open_ver_edit, mod)
                # NOTE: We have to keep a reference around for PySide.
                # PyQt4 handles these correctly...
                self._mvs_cbs.append((cb_sel, cb_edit))

                sel.currentIndexChanged.connect(cb_sel)
                editBut.clicked.connect(cb_edit)
            else:
                # TODO: Finish code for pinning dependencies.
                sel.setEnabled(False)
                editBut.setEnabled(False)

        layout.addItem(QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding))

    def update_versions(self, sel, mod, idx):
        version = sel.itemData(idx)

        if mod == self._mod:
            if version is None:
                center.installed.unpin(self._mod)
            else:
                if isinstance(version, str):
                    version = semantic_version.Version(version)

                center.installed.pin(self._mod, version)

            api.save_settings()
        else:
            raise NotImplemented

    def open_ver_edit(self, mod):
        ModVersionsWindow(mod)

    def check_files(self):
        self.win.setCursor(QtCore.Qt.BusyCursor)

        task = CheckFilesTask(self._mod)
        task.done.connect(functools.partial(self.__check_files, task))
        run_task(task)

    def __check_files(self, task):
        report = ''

        for pkg, s, c, info in task.get_results():
            if pkg is None:
                if len(info['loose']) > 0:
                    report += '<h2>Loose files (%d)</h2>' % (len(info['loose']))
                    report += '<ul><li>' + '</li><li>'.join(sorted(info['loose'])) + '</li></ul>'
            else:
                report += '<h2>%s (%d/%d files OK)</h2>' % (pkg.name, s, c)
                ok_count = len(info['ok'])
                corr_count = len(info['corrupt'])
                miss_count = len(info['missing'])

                report += '<ul>'

                if ok_count > 0:
                    if corr_count > 0 or miss_count > 0:
                        report += '<ul><li>OK<ul>'
                        report += '<li>' + '</li><li>'.join(sorted(info['ok'])) + '</li>'
                        report += '</ul></li>'
                    else:
                        report += '<li>' + '</li><li>'.join(sorted(info['ok'])) + '</li>'

                if corr_count > 0:
                    report += '<li>Corrupted<ul>'
                    report += '<li>' + '</li><li>'.join(sorted(info['corrupt'])) + '</li>'
                    report += '</ul></li>'

                if miss_count > 0:
                    report += '<li>Missing<ul>'
                    report += '<li>' + '</li><li>'.join(sorted(info['missing'])) + '</li>'
                    report += '</ul></li>'

                report += '</ul>'

        self.win.logDisplay.setHtml(report)
        self.win.unsetCursor()

    def delete_loose_files(self):
        self.win.setCursor(QtCore.Qt.BusyCursor)

        task = CheckFilesTask(self._mod)
        task.done.connect(functools.partial(self.__delete_loose_files, task))
        run_task(task)

    # TODO: Should this be moved into a task?
    def __delete_loose_files(self, task):
        modpath = os.path.join(center.settings['fs2_path'], self._mod.folder)

        for pkg, s, c, info in task.get_results():
            if pkg is None:
                for name in info['loose']:
                    item = os.path.join(modpath, name)
                    logging.info('Deleteing "%s"...', item)
                    os.unlink(item)

        self.win.unsetCursor()
        QtWidgets.QMessageBox.information(None, 'Knossos', 'Done!')

    def repair_files(self):
        try:
            run_task(InstallTask(self._mod.packages))
        except repo.ModNotFound as exc:
            QtWidgets.QMessageBox.critical(None, 'Knossos', "I can't repair this mod: %s" % str(exc))


class ModVersionsWindow(Window):
    _mod = None
    _versions = None

    def __init__(self, mod, window=True):
        super(ModVersionsWindow, self).__init__(window)

        self._mod = mod
        self._versions = []
        self._create_win(Ui_Mod_Versions, QDialog)
        self.win.setModal(True)

        self.label_tpl(self.win.label, MOD=mod.title)

        mods = list(center.mods.query_all(mod.mid))
        installed = list(center.installed.query_all(mod.mid))

        inst_versions = [m.version for m in installed]
        rem_versions = [m.version for m in mods]
        local_versions = set(inst_versions) - set(rem_versions)

        # Add all local-only versions
        for m in installed:
            if m.version in local_versions:
                mods.append(m)

        for m in mods:
            if len(m.packages) == 0:
                logging.warning('Version "%s" for mod "%s" (%s) is empty! (It has no packages!!)', m.version, m.title, m.mid)
                continue

            label = str(m.version)
            if m.version in local_versions:
                label += ' (l)'

            item = QtWidgets.QListWidgetItem(label, self.win.versionList)
            item.setData(QtCore.Qt.UserRole + 1, m)

            if m.version in local_versions:
                item.setToolTip('This version is installed locally but not available anymore!')

            if m.version in inst_versions:
                item.setCheckState(QtCore.Qt.Checked)
                item.setData(QtCore.Qt.UserRole, True)
            else:
                item.setCheckState(QtCore.Qt.Unchecked)
                item.setData(QtCore.Qt.UserRole, False)

            self._versions.append(item)

        self.win.applyButton.clicked.connect(self.apply_changes)
        self.win.cancelButton.clicked.connect(self.close)
        self.open()

    def apply_changes(self):
        install = set()
        uninstall = set()

        for item in self._versions:
            was = item.data(QtCore.Qt.UserRole)
            mod = item.data(QtCore.Qt.UserRole + 1)
            checked = item.checkState() == QtCore.Qt.Checked

            if was and not checked:
                uninstall |= set(center.installed.query(mod).packages)
            elif not was and checked:
                install |= set(mod.resolve_deps())

        install = center.mods.process_pkg_selection(install)

        if len(install) > 0:
            run_task(InstallTask(install, self._mod))

        if len(uninstall) > 0:
            run_task(UninstallTask(uninstall))

        self.close()


class ModInfoWindow(Window):
    _mod = None

    def __init__(self, mod, window=True):
        super(ModInfoWindow, self).__init__(window)

        self._mod = mod
        self._create_win(Ui_Mod_Settings, QDialog)

        self.win.modTitle.setText(mod.title)
        self.win.modLogo.setPixmap(QtWidgets.QPixmap(mod.logo))
        self.win.modDesc.setPlainText(mod.description)

        self.win.flagsTab.deleteLater()
        self.win.pkgTab.deleteLater()
        self.win.versionsTab.deleteLater()
        self.win.troubleTab.deleteLater()
        self.win.tabWidget.setCurrentIndex(0)

        self.open()


class LogViewer(Window):

    def __init__(self, path, window=True):
        super(LogViewer, self).__init__(window)

        self._create_win(Ui_Log_Viewer)
        self.win.pathLabel.setText(path)

        if not os.path.isfile(path):
            QtWidgets.QMessageBox.critical(None, 'Knossos', 'Log file %s can\'t be shown because it\'s missing!' % path)
            return

        with open(path, 'r') as stream:
            self.win.content.setPlainText(stream.read())

        self.open()



# This file mostly contains small parts of code from Python 3.4's stdlib.
# The code is used to make functions where were added or improved in Python 3
# available to Python 2 installations.

# The copied code has been slightly changed to make sure that all references
# work. However, the code's function has not been changed.

# The copied code is licensed under the PSF2 which is embedded below.

# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------

# 1. This LICENSE AGREEMENT is between the Python Software Foundation
# ("PSF"), and the Individual or Organization ("Licensee") accessing and
# otherwise using this software ("Python") in source or binary form and
# its associated documentation.

# 2. Subject to the terms and conditions of this License Agreement, PSF hereby
# grants Licensee a nonexclusive, royalty-free, world-wide license to reproduce,
# analyze, test, perform and/or display publicly, prepare derivative works,
# distribute, and otherwise use Python alone or in any derivative version,
# provided, however, that PSF's License Agreement and PSF's notice of copyright,
# i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
# 2011, 2012, 2013, 2014 Python Software Foundation; All Rights Reserved" are
# retained in Python alone or in any derivative version prepared by Licensee.

# 3. In the event Licensee prepares a derivative work that is based on
# or incorporates Python or any part thereof, and wants to make
# the derivative work available to others as provided herein, then
# Licensee hereby agrees to include in any such work a brief summary of
# the changes made to Python.

# 4. PSF is making Python available to Licensee on an "AS IS"
# basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND
# DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS
# FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT
# INFRINGE ANY THIRD PARTY RIGHTS.

# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON
# FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS
# A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON,
# OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.

# 6. This License Agreement will automatically terminate upon a material
# breach of its terms and conditions.

# 7. Nothing in this License Agreement shall be deemed to create any
# relationship of agency, partnership, or joint venture between PSF and
# Licensee.  This License Agreement does not grant permission to use PSF
# trademarks or trade name in a trademark sense to endorse or promote
# products or services of Licensee, or any third party.

# 8. By copying, installing or otherwise using Python, Licensee
# agrees to be bound by the terms and conditions of this License
# Agreement.

from __future__ import absolute_import, print_function

import os as _os
import warnings as _warnings
import tempfile
import subprocess
import re
import shlex
import shutil
import six

if six.PY3:
    raise RuntimeError('py2_compat was imported even though we are running python 3!!!')


# This class is copied straight from /usr/lib/python3.3/tempfile.py
class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix="", prefix="tmp", dir=None):
        self._closed = False
        self.name = None  # Handle mkdtemp raising an exception
        self.name = tempfile.mkdtemp(suffix, prefix, dir)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def cleanup(self, _warn=False):
        if self.name and not self._closed:
            try:
                self._rmtree(self.name)
            except (TypeError, AttributeError) as ex:
                # Issue #10188: Emit a warning on stderr
                # if the directory could not be cleaned
                # up due to missing globals
                if "None" not in str(ex):
                    raise
                self._warn("ERROR: {!r} while cleaning up {!r}".format(ex, self,), ResourceWarning)
                return
            self._closed = True
            if _warn:
                self._warn("Implicitly cleaning up {!r}".format(self),
                           ResourceWarning)

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def __del__(self):
        # Issue a ResourceWarning if implicit cleanup needed
        self.cleanup(_warn=True)

    # XXX (ncoghlan): The following code attempts to make
    # this class tolerant of the module nulling out process
    # that happens during CPython interpreter shutdown
    # Alas, it doesn't actually manage it. See issue #10188
    _listdir = staticmethod(_os.listdir)
    _path_join = staticmethod(_os.path.join)
    _isdir = staticmethod(_os.path.isdir)
    _islink = staticmethod(_os.path.islink)
    _remove = staticmethod(_os.remove)
    _rmdir = staticmethod(_os.rmdir)
    _os_error = OSError
    _warn = _warnings.warn

    def _rmtree(self, path):
        # Essentially a stripped down version of shutil.rmtree.  We can't
        # use globals because they may be None'ed out at shutdown.
        for name in self._listdir(path):
            fullname = self._path_join(path, name)
            try:
                isdir = self._isdir(fullname) and not self._islink(fullname)
            except self._os_error:
                isdir = False
            if isdir:
                self._rmtree(fullname)
            else:
                try:
                    self._remove(fullname)
                except self._os_error:
                    pass
        try:
            self._rmdir(path)
        except self._os_error:
            pass


# This code is copied straight from /usr/lib/python3.3/shlex.py
shlex_find_unsafe = re.compile(r'[^\w@%+=:,./-]').search


def shlex_quote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"
    if shlex_find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


# This code is copied straight from /usr/lib/python3.4/shutil.py
# Python >= 2.3 already had this function but in Python 3 support for symlinks was added.
def shutil_move(src, dst):
    """Recursively move a file or directory to another location. This is
    similar to the Unix "mv" command. Return the file or directory's
    destination.

    If the destination is a directory or a symlink to a directory, the source
    is moved inside the directory. The destination path must not already
    exist.

    If the destination already exists but is not a directory, it may be
    overwritten depending on os.rename() semantics.

    If the destination is on our current filesystem, then rename() is used.
    Otherwise, src is copied to the destination and then removed. Symlinks are
    recreated under the new name if os.rename() fails because of cross
    filesystem renames.

    A lot more could be done here...  A look at a mv.c shows a lot of
    the issues this implementation glosses over.

    """
    real_dst = dst
    if _os.path.isdir(dst):
        if shutil._samefile(src, dst):
            # We might be on a case insensitive filesystem,
            # perform the rename anyway.
            _os.rename(src, dst)
            return

        real_dst = _os.path.join(dst, shutil._basename(src))
        if _os.path.exists(real_dst):
            raise shutil.Error("Destination path '%s' already exists" % real_dst)
    try:
        _os.rename(src, real_dst)
    except OSError:
        if _os.path.islink(src):
            linkto = _os.readlink(src)
            _os.symlink(linkto, real_dst)
            _os.unlink(src)
        elif _os.path.isdir(src):
            if shutil._destinsrc(src, dst):
                raise shutil.Error("Cannot move a directory '%s' into itself '%s'." % (src, dst))
            shutil.copytree(src, real_dst, symlinks=True)
            shutil.rmtree(src)
        else:
            shutil.copy2(src, real_dst)
            _os.unlink(src)
    return real_dst


tempfile.TemporaryDirectory = TemporaryDirectory
shlex.quote = shlex_quote
shutil.move = shutil_move
subprocess.DEVNULL = open(_os.devnull, 'wb')



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import logging


# ugly, hacky fix
def uhf(n):
    np = n.split('.')
    if np[0] == __name__:
        globals()[np[1]] = sys.modules[n]
    else:
        logging.error('I can\'t add module %s to %s!', n, __name__)



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import os
import uuid
import logging
import functools

from .qt import QtCore, QtNetwork


class IPCComm(QtCore.QObject):
    _path = None
    _server = None
    _conn_pool = None
    _socket = None
    _queue = None
    messageReceived = QtCore.Signal(QtCore.QByteArray)

    def __init__(self, path):
        super(IPCComm, self).__init__()

        self._path = path
        self._queue = []

    def get_file(self):
        return os.path.join(self._path, 'ipcinfo')

    def server_exists(self):
        # TODO: Maybe we should check if the server is running?
        return os.path.isfile(self.get_file())

    def listen(self):
        if self._server is not None:
            raise Exception('IPCComm() is already listening!')

        name = 'knossos-' + str(uuid.uuid4())
        self._server = QtNetwork.QLocalServer()
        if not self._server.listen(name):
            msg = self._server.errorString()
            self._server = None
            raise Exception(msg)

        self._conn_pool = []
        self._server.newConnection.connect(self._accept)

        with open(self.get_file(), 'w') as stream:
            stream.write(self._server.fullServerName())

    # NOTE: connect() is already taken by Qt
    def open_connection(self, err_hdl=None):
        fp = self.get_file()
        if not os.path.isfile(fp):
            raise Exception('There is no server listening!')

        self._socket = QtNetwork.QLocalSocket()
        self._socket.error.connect(self._sock_error)
        self._socket.connected.connect(self._connected)

        if err_hdl is not None:
            self._socket.error.connect(err_hdl)

        with open(fp, 'r') as stream:
            addr = stream.read()

        self._socket.connectToServer(addr)

    def send_message(self, msg):
        if self._socket is None:
            self.open_connection()

        if self._socket.state() != QtNetwork.QLocalSocket.ConnectedState:
            self._queue.append(msg)
        else:
            self._socket.write(msg + '\n')

    def close(self, wait=False):
        if self._server is not None:
            self._server.close()
            self._server = None

            self.clean()

        elif self._socket is not None:
            if wait:
                self._socket.waitForBytesWritten(-1)

            self._socket.close()
            self._socket = None

        else:
            raise Exception('IPCComm() is neither listening nor open!')

    def clean(self):
        # NOTE: This disables the server!

        fp = self.get_file()
        if os.path.isfile(fp):
            os.unlink(fp)

    def _accept(self):
        conn = self._server.nextPendingConnection()
        conn.disconnected.connect(functools.partial(self._conn_disc, conn))
        conn.readyRead.connect(functools.partial(self._read_data, conn))

        self._conn_pool.append(conn)

    def _conn_disc(self, conn):
        self._conn_pool.remove(conn)

    def _read_data(self, conn):
        self.messageReceived.emit(conn.readLine())

    def _connected(self):
        for msg in self._queue:
            self._socket.write(msg + '\n')

        self._queue = []

    def _sock_error(self):
        logging.error(self._socket.errorString())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/install.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(460, 560)
        Dialog.setModal(True)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.splitter = QtWidgets.QSplitter(Dialog)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.verticalLayoutWidget = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.titleLabel = QtWidgets.QLabel(self.verticalLayoutWidget)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(True)
        font.setWeight(75)
        self.titleLabel.setFont(font)
        self.titleLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.titleLabel.setWordWrap(True)
        self.titleLabel.setObjectName("titleLabel")
        self.verticalLayout.addWidget(self.titleLabel)
        spacerItem = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem)
        self.treeWidget = QtWidgets.QTreeWidget(self.verticalLayoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.treeWidget.sizePolicy().hasHeightForWidth())
        self.treeWidget.setSizePolicy(sizePolicy)
        self.treeWidget.setAlternatingRowColors(True)
        self.treeWidget.setAnimated(True)
        self.treeWidget.setWordWrap(True)
        self.treeWidget.setObjectName("treeWidget")
        self.treeWidget.header().setDefaultSectionSize(300)
        self.verticalLayout.addWidget(self.treeWidget)
        self.dlSizeLabel = QtWidgets.QLabel(self.verticalLayoutWidget)
        self.dlSizeLabel.setObjectName("dlSizeLabel")
        self.verticalLayout.addWidget(self.dlSizeLabel)
        self.verticalLayoutWidget_2 = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget_2.setObjectName("verticalLayoutWidget_2")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.verticalLayoutWidget_2)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(self.verticalLayoutWidget_2)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.notesField = QtWidgets.QTextEdit(self.verticalLayoutWidget_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.notesField.sizePolicy().hasHeightForWidth())
        self.notesField.setSizePolicy(sizePolicy)
        self.notesField.setReadOnly(True)
        self.notesField.setObjectName("notesField")
        self.verticalLayout_2.addWidget(self.notesField)
        self.buttonBox = QtWidgets.QDialogButtonBox(self.verticalLayoutWidget_2)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_2.addWidget(self.buttonBox)
        self.verticalLayout_3.addWidget(self.splitter)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Install a mod"))
        self.titleLabel.setText(_translate("Dialog", "Install {MOD}?"))
        self.treeWidget.headerItem().setText(0, _translate("Dialog", "Name"))
        self.treeWidget.headerItem().setText(1, _translate("Dialog", "Filesize"))
        self.dlSizeLabel.setText(_translate("Dialog", "Download size: {DL_SIZE}"))
        self.label.setText(_translate("Dialog", "Notes:"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings_network.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.formLayout = QtWidgets.QFormLayout(Form)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.label_3 = QtWidgets.QLabel(Form)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.label_4 = QtWidgets.QLabel(Form)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.connectionType = QtWidgets.QComboBox(Form)
        self.connectionType.setObjectName("connectionType")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.connectionType)
        self.connectionSpeed = QtWidgets.QComboBox(Form)
        self.connectionSpeed.setObjectName("connectionSpeed")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.connectionSpeed)
        self.localPort = QtWidgets.QLineEdit(Form)
        self.localPort.setObjectName("localPort")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.localPort)
        self.forceAddress = QtWidgets.QLineEdit(Form)
        self.forceAddress.setObjectName("forceAddress")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.forceAddress)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "Connection type: "))
        self.label_2.setText(_translate("Form", "Connection speed: "))
        self.label_3.setText(_translate("Form", "Force local port: "))
        self.label_4.setText(_translate("Form", "Force IP address: "))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings_video.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.formLayout = QtWidgets.QFormLayout(Form)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.resolution = QtWidgets.QComboBox(Form)
        self.resolution.setObjectName("resolution")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.resolution)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.colorDepth = QtWidgets.QComboBox(Form)
        self.colorDepth.setObjectName("colorDepth")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.colorDepth)
        self.label_3 = QtWidgets.QLabel(Form)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.label_4 = QtWidgets.QLabel(Form)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.label_5 = QtWidgets.QLabel(Form)
        self.label_5.setObjectName("label_5")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.label_5)
        self.textureFilter = QtWidgets.QComboBox(Form)
        self.textureFilter.setObjectName("textureFilter")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.textureFilter)
        self.antialiasing = QtWidgets.QComboBox(Form)
        self.antialiasing.setObjectName("antialiasing")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.antialiasing)
        self.anisotropic = QtWidgets.QComboBox(Form)
        self.anisotropic.setObjectName("anisotropic")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.anisotropic)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "Resolution: "))
        self.label_2.setText(_translate("Form", "Color depth: "))
        self.label_3.setText(_translate("Form", "Texture filter: "))
        self.label_4.setText(_translate("Form", "Antialiasing: "))
        self.label_5.setText(_translate("Form", "Anisotropic filtering: "))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/settings_sources.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.sourceList = QtWidgets.QListWidget(Form)
        self.sourceList.setDragEnabled(True)
        self.sourceList.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.sourceList.setObjectName("sourceList")
        self.verticalLayout.addWidget(self.sourceList)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.addSource = QtWidgets.QPushButton(Form)
        self.addSource.setObjectName("addSource")
        self.horizontalLayout.addWidget(self.addSource)
        self.editSource = QtWidgets.QPushButton(Form)
        self.editSource.setObjectName("editSource")
        self.horizontalLayout.addWidget(self.editSource)
        self.removeSource = QtWidgets.QPushButton(Form)
        self.removeSource.setObjectName("removeSource")
        self.horizontalLayout.addWidget(self.removeSource)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.addSource.setText(_translate("Form", "Add"))
        self.editSource.setText(_translate("Form", "Edit"))
        self.removeSource.setText(_translate("Form", "Remove"))




## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import os
import sys
import logging
import subprocess
import shutil
import glob
import stat
import json
import tempfile
import threading
import random
import semantic_version

from . import center, util, progress, repo, api
from .repo import Repo
from .qt import QtGui


class FetchTask(progress.Task):

    def __init__(self):
        super(FetchTask, self).__init__()
        self.title = 'Fetching mod list...'

        # Remove all logos.
        for path in glob.glob(os.path.join(center.settings_path, 'logo*.*')):
            if os.path.isfile(path):
                logging.info('Removing old logo "%s"...', path)
                os.unlink(path)

        self.done.connect(self.finish)

        if repo.CPU_INFO is None:
            self.add_work([('init',)])
        else:
            self.add_work([('repo', i * 100, link[0]) for i, link in enumerate(center.settings['repos'])])

    def work(self, params):
        if params[0] == 'init':
            progress.update(0, 'Checking CPU...')

            # We're doing this here because we don't want to block the UI.
            repo.CPU_INFO = util.get_cpuinfo()

            self.add_work([('repo', i * 100, link[0]) for i, link in enumerate(center.settings['repos'])])
            return
        elif params[0] == 'repo':
            _, prio, link = params

            progress.update(0.1, 'Fetching "%s"...' % link)

            try:
                raw_data = util.get(link, raw=True)

                data = Repo()
                data.is_link = True
                data.base = os.path.dirname(raw_data.url)
                data.parse(raw_data.text)
            except:
                logging.exception('Failed to decode "%s"!', link)
                return

            wl = []
            for mid, mvs in data.mods.items():
                for mod in mvs:
                    wl.append(('mod', mod))

            self.add_work(wl)
            self.post((prio, data))
        else:
            mod = params[1]
            mod.save_logo(center.settings_path)

    def finish(self):
        if not self.aborted:
            modlist = center.mods = Repo()
            res = self.get_results()
            res.sort(key=lambda x: x[0])

            for part in res:
                modlist.merge(part[1])

            modlist.save_json(os.path.join(center.settings_path, 'mods.json'))
            api.save_settings()

        run_task(CheckTask())


class CheckTask(progress.MultistepTask):
    can_abort = False
    _steps = 2

    def __init__(self):
        super(CheckTask, self).__init__(threads=3)

        self.done.connect(self.finish)
        self.title = 'Checking installed mods...'

    def init1(self):
        if center.settings['fs2_path'] is None:
            logging.error('A CheckTask was launched even though no FS2 path was set!')
        else:
            center.installed.clear()
            self.add_work(('',))

    def work1(self, p):
        fs2path = center.settings['fs2_path']
        mods = center.installed

        for subdir in os.listdir(fs2path) + [fs2path]:
            kfile = os.path.join(fs2path, subdir, 'mod.')

            if os.path.isfile(kfile + 'json'):
                kfile = kfile + 'json'
            elif os.path.isfile(kfile + 'ini'):
                kfile = kfile + 'ini'
            else:
                continue

            try:
                mods.add_mod(repo.InstalledMod.load(kfile))
            except:
                logging.exception('Failed to parse "%s"!', kfile)
                continue

    def init2(self):
        pkgs = []
        for mid, mvs in center.installed.mods.items():
            for mod in mvs:
                pkgs.extend(mod.packages)

        # Reset them
        for pkg in pkgs:
            pkg.files_ok = 0
            pkg.files_checked = 0

        self.add_work(pkgs)

    def work2(self, pkg):
        fs2path = center.settings['fs2_path']
        mod = pkg.get_mod()
        modpath = os.path.join(fs2path, mod.folder)
        pkg_files = pkg.filelist
        count = float(len(pkg_files))
        success = 0
        missing = 0
        checked = 0

        archives = set()
        msgs = []

        for info in pkg_files:
            mypath = util.ipath(os.path.join(modpath, info['filename']))
            fix = False
            if os.path.isfile(mypath):
                progress.update(checked / count, 'Checking "%s"...' % (info['filename']))

                if util.gen_hash(mypath) == info['md5sum']:
                    success += 1
                else:
                    msgs.append('File "%s" is corrupted. (checksum mismatch)' % (info['filename']))
                    fix = True
            else:
                msgs.append('File "%s" is missing.' % (info['filename']))
                missing += 1
                fix = True

            if fix:
                archives.add(info['archive'])

            checked += 1

        self.post((pkg, archives, success, missing, checked, msgs))

    def finish(self):
        results = self.get_results()

        for pkg, archives, s, m, c, msg in results:
            mod = pkg.get_mod()

            if s == 0:
                if m > 0:
                    # Not Installed
                    # What?!
                    logging.warning('Package %s of mod %s (%s) is not installed but in the local repo. Fixing...' % (pkg.name, mod.mid, mod.title))
                    mod.del_pkg(pkg)
                    mod.save()
                elif c > 0:
                    logging.warning('Package %s of mod %s (%s) is completely corrupted!!' % (pkg.name, mod.mid, mod.title))

            if pkg is not None:
                pkg.check_notes = m
                pkg.files_ok = s
                pkg.files_checked = c

        center.signals.repo_updated.emit()


class CheckFilesTask(progress.MultistepTask):
    can_abort = False
    _mod = None
    _check_results = None
    _steps = 2

    def __init__(self, mod):
        super(CheckFilesTask, self).__init__(threads=3)

        self.title = 'Checking "%s"...' % mod.title

        if center.settings['fs2_path'] is None:
            logging.error('A CheckFilesTask was launched even though no FS2 path was set!')
            return

        self._mod = mod

    def init1(self):
        modpath = os.path.join(center.settings['fs2_path'], self._mod.folder)
        pkgs = []

        for pkg in self._mod.packages:
            pkgs.append((modpath, pkg))

        self.add_work(pkgs)

    def work1(self, data):
        modpath, pkg = data
        pkg_files = pkg.filelist
        count = float(len(pkg_files))
        success = 0
        checked = 0

        summary = {
            'ok': [],
            'corrupt': [],
            'missing': []
        }

        for info in pkg_files:
            mypath = util.ipath(os.path.join(modpath, info['filename']))
            if os.path.isfile(mypath):
                progress.update(checked / count, 'Checking "%s"...' % (info['filename']))

                if util.gen_hash(mypath) == info['md5sum']:
                    success += 1
                    summary['ok'].append(info['filename'])
                else:
                    summary['corrupt'].append(info['filename'])
            else:
                summary['missing'].append(info['filename'])

            checked += 1

        self.post((pkg, success, checked, summary))

    def init2(self):
        # Save the results from step 1.
        self._check_results = self.get_results()

        self.add_work(('',))

    def work2(self, d):
        modpath = os.path.join(center.settings['fs2_path'], self._mod.folder)
        fnames = set()
        loose = []

        # Collect all filenames
        for pkg, s, c, m in self._check_results:
            for info in pkg.filelist:
                fnames.add(info['filename'])

        # Ignore files are generated by Knossos
        fnames.add('mod.json')
        fnames.add('knossos.bmp')

        # Check for loose files.
        for path, dirs, files in os.walk(modpath):
            if path == modpath:
                subpath = ''
            else:
                subpath = os.path.relpath(path, modpath)

            for item in files:
                name = os.path.join(subpath, item).replace('\\', '/')
                if name not in fnames and not name.startswith('__k_plibs'):
                    loose.append(name)

        self._check_results.append((None, 0, 0, {'loose': loose}))
        self._results = self._check_results


# TODO: Optimize, make sure all paths are relative (no mod should be able to install to C:\evil)
# TODO: Add error messages.
class InstallTask(progress.MultistepTask):
    _pkgs = None
    _pkg_names = None
    _mods = None
    _dls = None
    _steps = 2
    _error = False
    _7z_lock = None
    mod = None
    check_after = True

    def __init__(self, pkgs, mod=None, check_after=True):
        self._pkgs = []
        self._pkg_names = []
        self.check_after = check_after

        if sys.platform == 'win32':
            self._7z_lock = threading.Lock()

        rmods = center.mods

        if mod is not None:
            self.mod = mod

        # Make sure we have remote mods here!
        for pkg in pkgs:
            mod = pkg.get_mod()
            pkg = rmods.query(pkg)
            self._pkgs.append(pkg)
            self._pkg_names.append((mod.mid, pkg.name))

        super(InstallTask, self).__init__()

        self.done.connect(self.finish)
        self.title = 'Installing mods...'

    def abort(self):
        super(InstallTask, self).abort()

        util.cancel_downloads()

    def finish(self):
        if self.aborted:
            if self._cur_step == 1:
                # Need to remove all those temporary directories.
                for ar in self.get_results():
                    shutil.rmtree(ar['tpath'])
        elif self._error:
            msg = (
                'An error occured during the installation of a mod. It might be partially installed.\n' +
                'If you need more help, ask ngld or read the debug log!'
            )
            QtGui.QMessageBox.critical(None, 'Knossos', msg)
        else:
            mods = set()

            # Register installed pkgs
            for pkg in self._pkgs:
                logo_path = pkg.get_mod().logo_path
                pkg = center.installed.add_pkg(pkg)
                mod = pkg.get_mod()
                mod.logo_path = logo_path
                mods.add(mod)

            # Generate mod.json files.
            for mod in mods:
                mod.save()

        # Revert all changes made to the mod object.
        for mod in self._mods:
            p = mod.folder.rfind('_kv_')
            if p > -1:
                mod.folder = mod.folder[:p]

        if self.check_after:
            run_task(CheckTask())

    def init1(self):
        mods = set()
        for pkg in self._pkgs:
            mods.add(pkg.get_mod())

        self._threads = 3
        self._mods = mods
        self.add_work(mods)

    def work1(self, mod):
        fs2_path = center.settings['fs2_path']
        modpath = os.path.join(fs2_path, mod.folder)
        mfiles = mod.get_files()
        mnames = [f['filename'] for f in mfiles] + ['knossos.bmp', 'mod.json']

        archives = set()
        progress.update(0, 'Checking %s...' % mod.title)

        kpath = os.path.join(modpath, 'mod.json')
        if os.path.isfile(kpath):
            try:
                with open(kpath, 'r') as stream:
                    info = json.load(stream)
            except:
                logging.exception('Failed to parse mod.json!')
                info = None

            if info is not None and info['version'] != str(mod.version):
                mod.folder += '_kv_' + str(mod.version)
                modpath = os.path.join(fs2_path, mod.folder)

        if mod.folder not in ('', '.') and os.path.isdir(modpath):
            for path, dirs, files in os.walk(modpath):
                relpath = path[len(modpath):].lstrip('/\\')
                for item in files:
                    itempath = util.pjoin(relpath, item)
                    if not itempath.startswith('__k_plibs') and itempath not in mnames:
                        logging.info('File "%s" is left over.', itempath)

        amount = float(len(mfiles))
        for i, info in enumerate(mfiles):
            if (mod.mid, info['package']) not in self._pkg_names:
                continue

            progress.update(i / amount, 'Checking %s: %s...' % (mod.title, info['filename']))

            itempath = util.ipath(os.path.join(modpath, info['filename']))
            if not os.path.isfile(itempath) or util.gen_hash(itempath) != info['md5sum']:
                archives.add((mod.mid, info['package'], info['archive']))

        self.post(archives)

    def init2(self):
        archives = set()
        downloads = []

        for a in self.get_results():
            archives |= a

        for pkg in self._pkgs:
            mod = pkg.get_mod()
            for item in pkg.files.values():
                if (mod.mid, pkg.name, item['filename']) in archives:
                    item = item.copy()
                    item['mod'] = mod
                    item['pkg'] = pkg
                    downloads.append(item)

        if len(archives) == 0:
            logging.info('Nothing to do for this InstallTask!')
        elif len(downloads) == 0:
            logging.error('Somehow we didn\'t find any downloads for this InstallTask!')
            self._error = True

        self._threads = 0
        self.add_work(downloads)

    def work2(self, archive):
        with tempfile.TemporaryDirectory() as tpath:
            arpath = os.path.join(tpath, archive['filename'])
            fs2path = center.settings['fs2_path']
            modpath = os.path.join(fs2path, archive['mod'].folder)

            # TODO: Maybe this should be an option?
            retries = 3
            done = False
            urls = list(archive['urls'])
            random.shuffle(urls)

            while retries > 0:
                retries -= 1

                for url in urls:
                    progress.start_task(0, 0.97, '%s')
                    progress.update(0, 'Ready')

                    with open(arpath, 'wb') as stream:
                        if not util.download(url, stream):
                            if self.aborted:
                                return

                            logging.error('Download of "%s" failed!', url)
                            continue

                    progress.finish_task()
                    progress.update(0.97, 'Checking "%s"...' % archive['filename'])

                    if util.gen_hash(arpath) == archive['md5sum']:
                        done = True
                        retries = 0
                        break
                    else:
                        logging.error('File "%s" is corrupted!', url)

            if not done:
                logging.error('Missing file "%s"!', archive['filename'])
                self._error = True
                return

            if self.aborted:
                return

            if archive['is_archive']:
                cpath = os.path.join(tpath, 'content')
                os.mkdir(cpath)

                needed_files = filter(lambda item: item['archive'] == archive['filename'], archive['pkg'].filelist)
                done = False

                if sys.platform == 'win32':
                    # Apparently I can't run multiple 7z instances on Windows. If I do, I always get the error
                    # "The archive can't be opened because it is still in use by another process."
                    # I have no idea why. It works fine on Linux and Mac OS.
                    # TODO: Is there a better solution?

                    progress.update(0.98, 'Waiting...')
                    self._7z_lock.acquire()

                progress.update(0.98, 'Extracting %s...' % archive['filename'])

                if util.extract_archive(arpath, cpath):
                    done = True
                    # Look for missing files
                    for item in needed_files:
                        src_path = os.path.join(cpath, item['orig_name'])

                        if not os.path.isfile(src_path):
                            logging.warning('Missing file "%s" from archive "%s" for package "%s" (%s)!',
                                            item['orig_name'], archive['filename'], archive['pkg'].name, archive['mod'].title)

                            done = False
                            break

                if sys.platform == 'win32':
                    self._7z_lock.release()

                if not done:
                    logging.error('Failed to unpack archive "%s" for package "%s" (%s)!',
                                  archive['filename'], archive['pkg'].name, archive['mod'].title)
                    shutil.rmtree(cpath, ignore_errors=True)
                    self._error = True
                    return

                for item in archive['pkg'].filelist:
                    if item['archive'] != archive['filename']:
                        continue

                    src_path = os.path.join(cpath, item['orig_name'])
                    dest_path = util.ipath(os.path.join(modpath, item['filename']))

                    try:
                        dparent = os.path.dirname(dest_path)
                        if not os.path.isdir(dparent):
                            os.makedirs(dparent)

                        shutil.move(src_path, dest_path)
                    except:
                        logging.exception('Failed to move file "%s" from archive "%s" for package "%s" (%s) to its destination %s!',
                                          src_path, archive['filename'], archive['pkg'].name, archive['mod'].title, dest_path)
                        self._error = True

                # Copy the remaining empty dirs and symlinks.
                for path, dirs, files in os.walk(cpath):
                    path = os.path.relpath(path, cpath)

                    for name in dirs:
                        src_path = os.path.join(cpath, path, name)
                        dest_path = util.ipath(os.path.join(modpath, path, name))

                        if os.path.islink(src_path):
                            if not os.path.lexists(dest_path):
                                linkto = os.readlink(src_path)
                                os.symlink(linkto, dest_path)
                        elif not os.path.exists(dest_path):
                            os.makedirs(dest_path)

                    for name in files:
                        src_path = os.path.join(cpath, path, name)

                        if os.path.islink(src_path):
                            dest_path = util.ipath(os.path.join(modpath, path, name))
                            if not os.path.lexists(dest_path):
                                linkto = os.readlink(src_path)
                                os.symlink(linkto, dest_path)
            else:
                for item in archive['pkg'].filelist:
                    if item['archive'] != archive['filename']:
                        continue

                    dest_path = util.ipath(os.path.join(modpath, archive['filename']))

                    try:
                        dparent = os.path.dirname(dest_path)
                        if not os.path.isdir(dparent):
                            os.makedirs(dparent)

                        shutil.move(arpath, dest_path)
                    except:
                        logging.exception('Failed to move file "%s" from archive "%s" for package "%s" (%s) to its destination %s!',
                                          arpath, archive['filename'], archive['pkg'].name, archive['mod'].title, dest_path)
                        self._error = True


# TODO: make sure all paths are relative (no mod should be able to install to C:\evil)
class UninstallTask(progress.MultistepTask):
    _pkgs = None
    _steps = 2
    check_after = True

    def __init__(self, pkgs, check_after=True):
        self._pkgs = []
        self.check_after = check_after

        for pkg in pkgs:
            try:
                self._pkgs.append(center.installed.query(pkg))
            except repo.ModNotFound:
                logging.exception('Someone tried to uninstall a non-existant package (%s, %s)! Skipping it...', pkg.get_mod().mid, pkg.name)

        super(UninstallTask, self).__init__()

        self.done.connect(self.finish)
        self.title = 'Uninstalling mods...'

    def init1(self):
        self.add_work(self._pkgs)

    def work1(self, pkg):
        fs2path = center.settings['fs2_path']
        mod = pkg.get_mod()

        for item in pkg.filelist:
            path = util.ipath(os.path.join(fs2path, mod.folder, item['filename']))
            if not os.path.isfile(path):
                logging.warning('File "%s" for mod "%s" (%s) is missing during uninstall!', item['filename'], mod.title, mod.mid)
            else:
                os.unlink(path)

    def init2(self):
        mods = set()

        # Unregister uninstalled pkgs.
        for pkg in self._pkgs:
            mods.add(pkg.get_mod())
            center.installed.del_pkg(pkg)

        self.add_work(mods)

    def work2(self, mod):
        modpath = os.path.join(center.settings['fs2_path'], mod.folder)

        if isinstance(mod, repo.IniMod):
            shutil.rmtree(modpath)
        elif len(mod.packages) == 0:
            # Remove our files

            my_files = (os.path.join(modpath, 'mod.json'), mod.logo_path)
            for path in my_files:
                if os.path.isfile(path):
                    os.unlink(path)

            libs = os.path.join(modpath, '__k_plibs')
            if os.path.isdir(libs):
                # Delete any symlinks before running shutil.rmtree().
                for link in os.listdir(libs):
                    item = os.path.join(libs, link)
                    if os.path.islink(item):
                        os.unlink(item)

                shutil.rmtree(libs)

            center.installed.del_mod(mod)
        else:
            mod.save()

        # Remove empty directories.
        for path, dirs, files in os.walk(modpath, topdown=False):
            if len(dirs) == 0 and len(files) == 0:
                os.rmdir(path)

    def finish(self):
        if self.check_after:
            run_task(CheckTask())


class UpdateTask(InstallTask):
    _old_mod = None
    _new_mod = None
    _new_modpath = None
    __check_after = True

    def __init__(self, mod, check_after=True):
        self._old_mod = mod
        self._new_mod = center.mods.query(mod.mid)
        self.__check_after = check_after

        old_pkgs = [pkg.name for pkg in mod.packages]
        pkgs = []

        for pkg in self._new_mod.packages:
            if pkg.name in old_pkgs or pkg.status == 'required':
                pkgs.append(pkg)

        super(UpdateTask, self).__init__(pkgs, self._new_mod, check_after=False)

    def finish(self):
        self._new_modpath = self._new_mod.folder
        super(UpdateTask, self).finish()

        if not self.aborted and not self._error:
            # TODO: Check if it's okay to uninstall the old version. Maybe there's still a mod that requires this one?
            # The new version has been succesfully installed, remove the old version.
            next_task = UninstallTask(self._old_mod.packages, check_after=False)
            next_task.done.connect(self._finish2)
            run_task(next_task)

    def _finish2(self):
        fs2path = center.settings['fs2_path']
        modpath = os.path.join(fs2path, self._old_mod.folder)
        temppath = os.path.join(fs2path, self._new_modpath)

        if '_kv_' not in modpath:
            # Move all files from the temporary directory to the new one.
            try:
                if not os.path.isdir(modpath):
                    os.makedirs(modpath)

                for item in os.listdir(temppath):
                    shutil.move(os.path.join(temppath, item), os.path.join(modpath, item))
            except:
                logging.exception('Failed to move a file!')
                QtGui.QMessageBox.critical(None, 'Knossos', 'Failed to replace the old mod files with the updated files!')

                if self.__check_after:
                    run_task(CheckTask())
                return

            try:
                # Remove the now empty temporary folder
                os.rmdir(temppath)
            except:
                logging.warning('Failed to remove supposedly empty folder "%s"!', temppath, exc_info=True)

        if self.__check_after:
            run_task(CheckTask())


class GOGExtractTask(progress.Task):
    can_abort = False

    def __init__(self, gog_path, dest_path):
        super(GOGExtractTask, self).__init__()

        self.done.connect(self.finish)
        self.add_work([(gog_path, dest_path)])
        self.title = 'Installing FS2 from GOG...'

    def work(self, paths):
        gog_path, dest_path = paths

        progress.update(0.03, 'Looking for InnoExtract...')
        data = util.get(center.INNOEXTRACT_LINK)

        try:
            data = json.loads(data)
        except:
            logging.exception('Failed to read JSON data!')
            return

        link = None
        path = None
        for plat, info in data.items():
            if sys.platform.startswith(plat):
                link, path = info
                break

        if link is None:
            logging.error('Couldn\'t find an innoextract download for "%s"!', sys.platform)
            return

        inno = os.path.join(dest_path, os.path.basename(path))
        with tempfile.TemporaryDirectory() as tempdir:
            archive = os.path.join(tempdir, os.path.basename(link))

            progress.start_task(0.03, 0.10, 'Downloading InnoExtract...')
            with open(os.path.join(dest_path, archive), 'wb') as dl:
                util.download(link, dl)

            progress.finish_task()
            progress.update(0.13, 'Extracting InnoExtract...')

            util.extract_archive(archive, tempdir)
            shutil.move(os.path.join(tempdir, path), inno)

        # Make it executable
        mode = os.stat(inno).st_mode
        os.chmod(inno, mode | stat.S_IXUSR)

        progress.start_task(0.15, 0.75, 'Extracting FS2: %s')
        try:
            cmd = [inno, '-L', '-s', '-p', '-e', gog_path]
            logging.info('Running %s...', ' '.join(cmd))

            opts = dict()
            if sys.platform.startswith('win'):
                si = subprocess.STARTUPINFO()
                si.dwFlags = subprocess.STARTF_USESHOWWINDOW
                si.wShowWindow = subprocess.SW_HIDE

                opts['startupinfo'] = si
                opts['stdin'] = subprocess.PIPE

            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=dest_path, **opts)

            if sys.platform.startswith('win'):
                p.stdin.close()

            buf = ''
            while p.poll() is None:
                while '\r' not in buf:
                    line = p.stdout.read(10)
                    if not line:
                        break
                    buf += line.decode('utf8', 'replace')

                buf = buf.split('\r')
                line = buf.pop(0)
                buf = '\r'.join(buf)

                if 'MiB/s' in line:
                    try:
                        if ']' in line:
                            line = line.split(']')[1]

                        line = line.strip().split('MiB/s')[0] + 'MiB/s'
                        percent = float(line.split('%')[0]) / 100

                        progress.update(percent, line)
                    except:
                        logging.exception('Failed to process InnoExtract output!')
                else:
                    if line.strip() == 'not a supported Inno Setup installer':
                        self.post(-1)
                        return

                    logging.info('InnoExtract: %s', line)
        except:
            logging.exception('InnoExtract failed!')
            return

        progress.finish_task()

        progress.update(0.95, 'Moving files...')
        self._makedirs(os.path.join(dest_path, 'data/players'))
        self._makedirs(os.path.join(dest_path, 'data/movies'))

        for item in glob.glob(os.path.join(dest_path, 'app', '*.vp')):
            shutil.move(item, os.path.join(dest_path, os.path.basename(item)))

        for item in glob.glob(os.path.join(dest_path, 'app/data/players', '*.hcf')):
            shutil.move(item, os.path.join(dest_path, 'data/players', os.path.basename(item)))

        for item in glob.glob(os.path.join(dest_path, 'app/data2', '*.mve')):
            shutil.move(item, os.path.join(dest_path, 'data/movies', os.path.basename(item)))

        for item in glob.glob(os.path.join(dest_path, 'app/data3', '*.mve')):
            shutil.move(item, os.path.join(dest_path, 'data/movies', os.path.basename(item)))

        progress.update(0.99, 'Cleanup...')
        os.unlink(inno)
        shutil.rmtree(os.path.join(dest_path, 'app'), ignore_errors=True)
        shutil.rmtree(os.path.join(dest_path, 'tmp'), ignore_errors=True)

        self.post(dest_path)

    def _makedirs(self, path):
        if not os.path.isdir(path):
            os.makedirs(path)

    def finish(self):
        results = self.get_results()
        if len(results) < 1:
            QtGui.QMessageBox.critical(None, 'Error', 'The installer failed! Please read the log for more details...')
            return
        elif results[0] == -1:
            QtGui.QMessageBox.critical(None, 'Error', 'The selected file wasn\'t a proper Inno Setup installer. Are you shure you selected the right file?')
            return
        else:
            QtGui.QMessageBox.information(None, 'Done', 'FS2 has been successfully installed.')

        fs2_path = results[0]
        center.settings['fs2_path'] = fs2_path
        center.settings['fs2_bin'] = None

        for item in glob.glob(os.path.join(fs2_path, 'fs2_*.exe')):
            if os.path.isfile(item):
                center.settings['fs2_bin'] = os.path.basename(item)
                break

        api.save_settings()
        center.main_win.check_fso()


class CheckUpdateTask(progress.Task):
    background = True

    def __init__(self):
        super(CheckUpdateTask, self).__init__()

        self.add_work(('',))
        self.title = 'Checking for updates...'

    def work(self, item):
        progress.update(0, 'Checking for updates...')

        update_base = util.pjoin(center.UPDATE_LINK, center.settings['update_channel'])
        version = util.get(update_base + '/version?me=' + center.VERSION)

        if version is None:
            logging.error('Update check failed!')
            return

        try:
            version = semantic_version.Version(version)
        except:
            logging.exception('Failed to parse remote version!')
            return

        cur_version = semantic_version.Version(center.VERSION)
        if version > cur_version:
            center.signals.update_avail.emit(version)


class WindowsUpdateTask(progress.Task):

    def __init__(self):
        super(WindowsUpdateTask, self).__init__()

        self.done.connect(self.finish)
        self.add_work(('',))
        self.title = 'Installing update...'

    def work(self, item):
        # Download it.
        update_base = util.pjoin(center.UPDATE_LINK, center.settings['update_channel'])

        dir_name = tempfile.mkdtemp()
        updater = os.path.join(dir_name, 'knossos_updater.exe')

        progress.start_task(0, 0.98, 'Downloading update...')
        with open(updater, 'wb') as stream:
            util.download(update_base + '/updater.exe', stream)

        progress.finish_task()
        progress.update(0.99, 'Launching updater...')

        try:
            import win32api
            win32api.ShellExecute(0, 'open', updater, '/D=' + os.getcwd(), os.path.dirname(updater), 1)
        except:
            logging.exception('Failed to launch updater!')
            self.post(False)
        else:
            self.post(True)
            center.app.quit()

    def finish(self):
        res = self.get_results()

        if len(res) < 1 or not res[0]:
            QtGui.QMessageBox.critical(None, 'Knossos', 'Failed to launch the update!')


def run_task(task, cb=None):
    def wrapper():
        cb(task.get_results())

    if cb is not None:
        task.done.connect(wrapper)

    center.pmaster.add_task(task)
    center.signals.task_launched.emit(task)
    return task



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging
import re
import threading
import subprocess
import time
import ctypes.util
import stat
import shutil
import six

from . import center, util
from .qt import QtCore, QtGui

# TODO: What happens if a SONAME contains a space?
FILE_PATH_RE = r'[a-zA-Z0-9\./\-\_\+]+'
LDD_RE = re.compile(r'\s*(' + FILE_PATH_RE + r') (?:=> (not found|' + FILE_PATH_RE + r'))?(?: \([^\)]+\))?\s*')
LIB_RE = re.compile(r'lib([a-zA-Z0-9\.\-\_\+]+)\.so(?:\..*)?')
LDCONF_RE = re.compile(r'\s*(' + FILE_PATH_RE + r') \([^\)]+\) => (' + FILE_PATH_RE + r')')
_LIB_CACHE = None


fs2_watcher = None


class SignalContainer(QtCore.QObject):
    signal = QtCore.Signal(list)


# This wrapper makes sure that the wrapped function is always run in the QT main thread.
def run_in_qt(func):
    cont = SignalContainer()
    
    def dispatcher(*args):
        cont.signal.emit(list(args))
    
    def listener(params):
        func(*params)
    
    cont.signal.connect(listener)
    
    return dispatcher


class Fs2Watcher(threading.Thread):
    _params = None
    _key_layout = None
    
    def __init__(self, params=None):
        super(Fs2Watcher, self).__init__()
        
        if params is None:
            self._params = []
        else:
            self._params = params
        
        self.daemon = True
        self.start()
    
    def run(self):
        global fs2_watcher

        old_path = None

        if center.settings['keyboard_layout'] != 'default':
            self._params.append('-keyboard_layout')
            self._params.append(center.settings['keyboard_layout'])
        
        fs2_bin = os.path.join(center.settings['fs2_path'], center.settings['fs2_bin'])
        if not os.path.isfile(fs2_bin):
            self.fs2_missing_msg(fs2_bin)
            return

        mode = os.stat(fs2_bin).st_mode
        if mode & stat.S_IXUSR != stat.S_IXUSR:
            # Make it executable.
            os.chmod(fs2_bin, mode | stat.S_IXUSR)

        env = os.environ.copy()
        if sys.platform.startswith('linux'):
            ld_path, missing = fix_missing_libs(fs2_bin)
            if len(missing) > 0:
                self.complain_missing(missing)
                return

            env['LD_LIBRARY_PATH'] = ld_path

        if center.settings['keyboard_setxkbmap']:
            self.set_us_layout()
        
        logging.debug('Launching FS2: %s', [fs2_bin] + self._params)

        if sys.platform.startswith('win'):
            bin_path = center.settings['fs2_bin']
            
            if os.path.basename(bin_path) != bin_path:
                # On Windows, the FSO engine changes the CWD to the directory the EXE file is in.
                # Since the fs2_bin is in a subdirectory we'll have to copy it!
                old_path = fs2_bin
                fs2_bin = os.path.join(center.settings['fs2_path'], '__tmp_' + os.path.basename(old_path))

                shutil.copy2(old_path, fs2_bin)

                # Make sure FSO still finds any DLL files located in the original EXE's folder.
                old_parent = os.path.abspath(os.path.dirname(old_path))
                if 'PATH' in env:
                    env['PATH'] = old_parent + os.pathsep + env['PATH']
                else:
                    env['PATH'] = old_parent

                if six.PY2:
                    env['PATH'] = env['PATH'].encode('utf8')

        fail = False
        rc = -999
        reason = '???'
        try:
            p = subprocess.Popen([fs2_bin] + self._params, cwd=center.settings['fs2_path'], env=env)

            time.sleep(0.3)
            if p.poll() is not None:
                rc = p.returncode
                reason = 'return code: %d' % rc
                fail = True
        except OSError as exc:
            logging.exception('Failed to launch FS2!')
            reason = str(exc).decode('utf8', 'replace')
            fail = True

        if fail:
            self.cleanup(fs2_bin, old_path)
            center.signals.fs2_failed.emit(rc)
            self.failed_msg(reason)
            return
        
        center.signals.fs2_launched.emit()
        p.wait()
        self.cleanup(fs2_bin, old_path)
        center.signals.fs2_quit.emit()
    
    @run_in_qt
    def failed_msg(self, reason):
        msg = 'Starting FS2 Open (%s) failed! (%s)' % (os.path.join(center.settings['fs2_path'], center.settings['fs2_bin']), reason)
        QtGui.QMessageBox.critical(center.app.activeWindow(), 'Failed', msg)

    @run_in_qt
    def fs2_missing_msg(self, fs2_bin):
        QtGui.QMessageBox.critical(None, 'Knossos', 'I can\'t find FSO! (The file "%s" is missing!)' % fs2_bin)

    @run_in_qt
    def complain_missing(self, missing):
        if len(missing) > 1:
            msg = "I can't start FSO because the libraries %s are missing!"
        else:
            msg = "I can't start FSO because the library %s is missing!"
        
        QtGui.QMessageBox.critical(None, 'Knossos', msg % util.human_list(missing))

    def set_us_layout(self):
        key_layout = util.check_output(['setxkbmap', '-query'])
        self._key_layout = key_layout.splitlines()[2].split(':')[1].strip()

        util.call(['setxkbmap', '-layout', 'us'])

    def cleanup(self, fs2_bin, old_path=None):
        if self._key_layout is not None:
            util.call(['setxkbmap', '-layout', self._key_layout])

        if sys.platform.startswith('win') and old_path is not None:
            # Cleanup
            retries = 3

            while retries > 0:
                try:
                    os.unlink(fs2_bin)
                except:
                    logging.exception('Failed to delete FSO file "%s"!', fs2_bin)

                if os.path.isfile(fs2_bin):
                    time.sleep(1)
                    retries -= 1
                else:
                    break


def check_elf_libs(fpath):
    out = util.check_output(['ldd', fpath], env={'LANG': 'C'}).splitlines()
    libs = {}

    for r_line in out:
        line = LDD_RE.match(r_line)
        if not line:
            logging.warning('Failed to parse ldd line "%s"!', r_line)
            continue

        if line.group(2) == 'not found':
            libs[line.group(1)] = None
        elif not line.group(2):
            # This is most likely a line like "linux-vdso.so.1 (0x00007fff847fe000)" which we can ignore.
            continue
        else:
            libs[line.group(1)] = line.group(2)

    return libs


def get_lib_path(filename):
    global _LIB_CACHE

    if not _LIB_CACHE:
        data = util.check_output(['ldconfig', '-p'], env={'LANG': 'C'}).splitlines()
        _LIB_CACHE = {}

        for line in data[1:]:
            m = LDCONF_RE.match(line)
            if not m:
                logging.warning('Failed to parse ldconfig line "%s"!', line)
                continue

            _LIB_CACHE[m.group(1)] = m.group(2)

    return _LIB_CACHE.get(filename)


def fix_missing_libs(fpath):
    base = os.path.dirname(fpath)
    patch_dir = os.path.join(base, '__k_plibs')

    libs = check_elf_libs(fpath)
    missing = []
    for name, path in libs.items():
        if not path:
            missing.append(name)

    if len(missing) == 0:
        # Yay, nothing to do.
        return '', missing

    for lib in missing[:]:
        p_name = os.path.join(patch_dir, lib)
        if os.path.exists(p_name):
            missing.remove(lib)
        else:
            ld_name = LIB_RE.match(lib)
            if ld_name:
                fixed_name = get_lib_path(ctypes.util.find_library(ld_name.group(1)))
                if fixed_name:
                    if not os.path.isdir(patch_dir):
                        os.mkdir(patch_dir)

                    os.symlink(fixed_name, p_name)
                    missing.remove(lib)

    return patch_dir, missing


def run_fs2(params=None):
    global fs2_watcher
    
    if fs2_watcher is None or not fs2_watcher.is_alive():
        fs2_watcher = Fs2Watcher(params)
        return True
    else:
        return False


def run_fs2_silent(params):
    fs2_path = center.settings['fs2_path']
    fs2_bin = os.path.join(fs2_path, center.settings['fs2_bin'])

    if not os.path.isfile(fs2_bin):
        return -128

    mode = os.stat(fs2_bin).st_mode
    if mode & stat.S_IXUSR != stat.S_IXUSR:
        # Make it executable.
        os.chmod(fs2_bin, mode | stat.S_IXUSR)

    env = {}
    if sys.platform.startswith('linux'):
        ld_path, missing = fix_missing_libs(fs2_bin)
        if len(missing) > 0:
            return -127

        env['LD_LIBRARY_PATH'] = ld_path

    try:
        return util.call([fs2_bin] + params, cwd=fs2_path, env=env)
    except OSError:
        return -129



## Copyright 2015 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging
import ctypes
import shlex
import six

from . import center, clibs, qt, launcher

if qt.variant == 'PySide':
    if six.PY2:
        ctypes.pythonapi.PyCObject_AsVoidPtr.argtypes = [ctypes.py_object]
        ctypes.pythonapi.PyCObject_AsVoidPtr.restype = ctypes.c_void_p
    else:
        ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object, ctypes.c_char_p]
        ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p


class Integration(object):

    def shutdown(self):
        pass

    def show_progress(self, value):
        pass

    def set_progress(self, value):
        pass

    def hide_progress(self):
        pass

    def annoy_user(self, activate=True):
        pass

    def install_scheme_handler(self):
        pass


class LinuxIntegration(Integration):

    def install_scheme_handler(self):
        my_cmd = launcher.get_cmd()

        tpl_desktop = r"""[Desktop Entry]
Name=Knossos
Exec={PATH} %U
Icon={ICON_PATH}
Type=Application
Terminal=false
MimeType=x-scheme-handler/fso;
"""

        tpl_mime_type = 'x-scheme-handler/fso=Knossos.desktop;'

        applications_path = os.path.expanduser('~/.local/share/applications/')
        desktop_file = applications_path + 'knossos.desktop'
        mime_types_file = applications_path + 'mimeapps.list'
        my_path = ' '.join([shlex.quote(p) for p in my_cmd])

        if os.path.isfile(desktop_file) or os.path.isfile('/usr/share/applications/knossos.desktop'):
            return True

        tpl_desktop = tpl_desktop.replace('{PATH}', my_path)
        tpl_desktop = tpl_desktop.replace('{ICON_PATH}', launcher.get_file_path('hlp.png'))

        if not os.path.isdir(applications_path):
            os.makedirs(applications_path)

        with open(desktop_file, 'w') as output_file:
            output_file.write(tpl_desktop)

        found = False
        if os.path.isfile(mime_types_file):
            with open(mime_types_file, 'r') as lines:
                for line in lines:
                    if tpl_mime_type in line:
                        found = True
                        break

        if not found:
            with open(mime_types_file, 'a') as output_file:
                output_file.write(tpl_mime_type)

        return True


class UnityIntegration(LinuxIntegration):
    launcher = None

    def __init__(self, Unity):
        self.launcher = Unity.LauncherEntry.get_for_desktop_id('knossos.desktop')

    def show_progress(self, value):
        self.launcher.set_property('progress_visible', True)
        self.set_progress(value)

    def set_progress(self, value):
        self.launcher.set_property('progress', value)

    def hide_progress(self):
        self.launcher.set_property('progress_visible', False)

    def annoy_user(self, activate=True):
        self.launcher.set_property('urgent', activate)


class WindowsIntegration(Integration):
    TBPF_NOPROGRESS = 0x0
    TBPF_INDETERMINATE = 0x1
    TBPF_NORMAL = 0x2
    TBPF_ERROR = 0x4
    TBPF_PAUSED = 0x8

    taskbar = None
    _hwnd = None
    _win = None

    def __init__(self, taskbar):
        self.taskbar = taskbar
        taskbar.HrInit()

    def wid(self):
        win = center.main_win.win
        if win != self._win:
            if qt.variant == 'PySide':
                if six.PY2:
                    self._hwnd = ctypes.pythonapi.PyCObject_AsVoidPtr(win.winId())
                else:
                    self._hwnd = ctypes.pythonapi.PyCapsule_GetPointer(win.winId(), None)
            else:
                self._hwnd = win.winId()
            self._win = win

        return self._hwnd

    def show_progress(self, value):
        self.taskbar.SetProgressState(self.wid(), self.TBPF_NORMAL)
        self.set_progress(value)

    def set_progress(self, value):
        self.taskbar.SetProgressValue(self.wid(), int(value * 100), 100)

    def hide_progress(self):
        self.taskbar.SetProgressState(self.wid(), self.TBPF_NOPROGRESS)

    def annoy_user(self, activate=True):
        pass

    def install_scheme_handler(self):
        my_cmd = launcher.get_cmd()

        settings = qt.QtCore.QSettings('HKEY_CLASSES_ROOT\\fso', qt.QtCore.QSettings.NativeFormat)
        settings.setFallbacksEnabled(False)

        settings.setValue('Default', 'URL:Knossos protocol')
        settings.setValue('URL Protocol', '')
        settings.setValue('DefaultIcon/Default', '"' + launcher.get_file_path('hlp.ico') + ',1"')

        my_cmd.append('%1')
        my_path = ' '.join(['"' + p + '"' for p in my_cmd])

        settings.setValue('shell/open/command/Default', my_path)

        # Check
        # FIXME: Is there any better way to detect whether this worked or not?

        settings.sync()
        settings = qt.QtCore.QSettings('HKEY_CLASSES_ROOT\\fso', qt.QtCore.QSettings.NativeFormat)
        settings.setFallbacksEnabled(False)

        return settings.value('shell/open/command/Default') == my_path

current = None


def init():
    global current

    if sys.platform.startswith('win'):
        try:
            import comtypes.client as cc

            tbl = cc.GetModule('taskbar.tlb')
            taskbar = cc.CreateObject('{56FDF344-FD6D-11d0-958A-006097C9A090}', interface=tbl.ITaskbarList3)
        except:
            logging.exception('Failed to load ITaskbarList3! Disabling Windows integration.')
        else:
            logging.info('Activating Windows integration...')
            current = WindowsIntegration(taskbar)
            return
    elif sys.platform.startswith('linux'):
        # TODO: Do we really have to initialize GTK?

        clibs.init_gtk()
        theme = clibs.get_gtk_theme()
        if theme == 'Ambiance':
            center.main_win.win.setStyleSheet(qt.load_styles(':/ui/themes/Ambiance/style.css'))
            center.main_win.win.setContentsMargins(-9, -9, -9, -9)
            center.main_win.win.verticalLayout.setSpacing(0)

        try:
            from gi.repository import Unity
        except ImportError:
            # Can't find Unity.
            pass
        else:
            logging.info('Activating Unity integration...')
            current = UnityIntegration(Unity)
            return

        logging.info('Activating generic Linux integration...')
        current = LinuxIntegration()
        return

    logging.warning('No desktop integration active.')
    current = Integration()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/gogextract.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(394, 167)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setWordWrap(True)
        self.label_3.setObjectName("label_3")
        self.verticalLayout.addWidget(self.label_3)
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.destPath = QtWidgets.QLineEdit(Dialog)
        self.destPath.setObjectName("destPath")
        self.gridLayout_2.addWidget(self.destPath, 6, 1, 1, 1)
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 3, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setObjectName("label_2")
        self.gridLayout_2.addWidget(self.label_2, 6, 0, 1, 1)
        self.gogPath = QtWidgets.QLineEdit(Dialog)
        self.gogPath.setObjectName("gogPath")
        self.gridLayout_2.addWidget(self.gogPath, 3, 1, 1, 1)
        self.gogButton = QtWidgets.QPushButton(Dialog)
        self.gogButton.setObjectName("gogButton")
        self.gridLayout_2.addWidget(self.gogButton, 3, 2, 1, 1)
        self.destButton = QtWidgets.QPushButton(Dialog)
        self.destButton.setObjectName("destButton")
        self.gridLayout_2.addWidget(self.destButton, 6, 2, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout_2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.cancelButton = QtWidgets.QPushButton(Dialog)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        self.installButton = QtWidgets.QPushButton(Dialog)
        self.installButton.setEnabled(False)
        self.installButton.setObjectName("installButton")
        self.horizontalLayout.addWidget(self.installButton)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Extract from GOG Installer"))
        self.label_3.setText(_translate("Dialog", "Please select the installer which you downloaded from GOG and a destination directory where FS2 should be installed."))
        self.label.setText(_translate("Dialog", "GOG Installer:"))
        self.label_2.setText(_translate("Dialog", "Destination:"))
        self.gogButton.setText(_translate("Dialog", "..."))
        self.destButton.setText(_translate("Dialog", "..."))
        self.cancelButton.setText(_translate("Dialog", "Cancel"))
        self.installButton.setText(_translate("Dialog", "Install"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/mod_versions.ui'
#
# Created by: PyQt5 UI code generator 5.7.1
#
# WARNING! All changes made in this file will be lost!

from ..qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(400, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.versionList = QtWidgets.QListWidget(Dialog)
        self.versionList.setObjectName("versionList")
        self.verticalLayout.addWidget(self.versionList)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.applyButton = QtWidgets.QPushButton(Dialog)
        self.applyButton.setObjectName("applyButton")
        self.horizontalLayout.addWidget(self.applyButton)
        self.cancelButton = QtWidgets.QPushButton(Dialog)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Mod versions"))
        self.label.setText(_translate("Dialog", "Available versions for {MOD}:"))
        self.applyButton.setText(_translate("Dialog", "Apply"))
        self.cancelButton.setText(_translate("Dialog", "Cancel"))

